#!/usr/bin/env python3

from argparse import ArgumentParser
from json import dump
from pathlib import Path

from pandas import concat, read_json


def parse_args():
    parser = ArgumentParser(description="Archive Query Log Demo")

    parser.add_argument(
        '--input',
        type=Path,
        help="The directory with the input data "
             "(i.e., a directory containing serps and results).",
        required=True,
    )
    parser.add_argument(
        "--output",
        type=Path,
        help="The output will be stored in this directory.",
        required=True,
    )

    return parser.parse_args()


def main(input_dir: Path, output_dir: Path):
    paths = input_dir.glob("serps/part*.gz")
    df = concat([read_json(path, lines=True) for path in paths])
    with (output_dir / "results.json").open("w") as file:
        dump(df["search_provider_name"].value_counts().to_dict(), file)


if __name__ == "__main__":
    args = parse_args()
    main(args.input, args.output)



from datetime import datetime
from pathlib import Path
from typing import Type

from elasticsearch_dsl.query import Exists, Query, Term
from expiringdict import ExpiringDict
from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from pydantic import BaseModel

from archive_query_log.config import Config
from archive_query_log.orm import (
    Archive,
    Provider,
    Source,
    Capture,
    BaseDocument,
    Serp,
    Result,
    UrlQueryParser,
    UrlPageParser,
    UrlOffsetParser,
    WarcQueryParser,
    WarcSnippetsParser,
)

_CACHE_SECONDS_STATISTICS = 60 * 5  # 5 minutes
_CACHE_SECONDS_PROGRESS = 60 * 10  # 10 minutes

_DEFAULT_CONFIG_PATH = Path("../../config.yml")
_DEFAULT_CONFIG_OVERRIDE_PATH = Path("../../config.override.yml")
_DEFAULT_CONFIG_PATHS = []
if _DEFAULT_CONFIG_PATH.exists():
    _DEFAULT_CONFIG_PATHS.append(_DEFAULT_CONFIG_PATH)
if _DEFAULT_CONFIG_OVERRIDE_PATH.exists():
    _DEFAULT_CONFIG_PATHS.append(_DEFAULT_CONFIG_OVERRIDE_PATH)


class Statistics(BaseModel):
    name: str
    description: str
    total: int
    disk_size: str | None
    last_modified: datetime | None


class Progress(BaseModel):
    input_name: str
    output_name: str
    description: str
    total: int
    current: int


DocumentType = Type[BaseDocument]

_statistics_cache: dict[
    tuple[DocumentType, str, str],
    Statistics,
] = ExpiringDict(
    max_len=100,
    max_age_seconds=_CACHE_SECONDS_STATISTICS,
)


def _convert_bytes(bytes_count: int) -> str:
    step_unit = 1000.0
    bytes_count_decimal: float = bytes_count
    for unit in ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB", "RB"]:
        if bytes_count_decimal < step_unit:
            return f"{bytes_count_decimal:3.1f} {unit}"
        bytes_count_decimal /= step_unit
    return f"{bytes_count_decimal:3.1f} QB"


def _get_statistics(
    config: Config,
    name: str,
    description: str,
    document: DocumentType,
    index: str,
    filter_query: Query | None = None,
) -> Statistics:
    key = (document, index, repr(filter_query))
    if key in _statistics_cache:
        return _statistics_cache[key]

    config.es.client.indices.refresh(index=index)
    stats = config.es.client.indices.stats(index=index)
    search = document.search(using=config.es.client, index=index)
    if filter_query is not None:
        search = search.filter(filter_query)
    total = search.count()
    last_modified_response = (
        search.query(Exists(field="last_modified"))
        .sort("-last_modified")
        .extra(size=1)
        .execute()
    )
    if last_modified_response.hits.total.value == 0:
        last_modified = None
    else:
        last_modified = last_modified_response.hits[0].last_modified

    statistics = Statistics(
        name=name,
        description=description,
        total=total,
        disk_size=str(
            _convert_bytes(stats["_all"]["total"]["store"]["size_in_bytes"])
            if filter_query is None
            else None
        ),
        last_modified=last_modified,
    )
    _statistics_cache[key] = statistics
    return statistics


_progress_cache: dict[
    tuple[DocumentType, str, str, str],
    Progress,
] = ExpiringDict(
    max_len=100,
    max_age_seconds=_CACHE_SECONDS_PROGRESS,
)


def _get_processed_progress(
    config: Config,
    input_name: str,
    output_name: str,
    description: str,
    document: DocumentType,
    index: str,
    status_field: str,
    filter_query: Query | None = None,
) -> Progress:
    key = (document, index, repr(filter_query), status_field)
    if key in _progress_cache:
        return _progress_cache[key]

    config.es.client.indices.refresh(index=index)
    search = document.search(using=config.es.client, index=index)
    if filter_query is not None:
        search = search.filter(filter_query)
    total = search.count()
    search_processed = search.filter(Term(**{status_field: False}))
    total_processed = search_processed.count()
    progress = Progress(
        input_name=input_name,
        output_name=output_name,
        description=description,
        total=total,
        current=total_processed,
    )
    _progress_cache[key] = progress
    return progress


# how to start api in bash: "uvicorn main:app --reload"

app = FastAPI()

# CORS settings
origins = [
    "http://localhost:3000",
]

app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
)


@app.get("/statistics")
def get_statistics() -> list[Statistics]:
    config: Config = Config()

    statistics_list = [
        _get_statistics(
            config=config,
            name="Archives",
            description="Web archiving services that offer CDX and Memento APIs.",
            document=Archive,
            index=config.es.index_archives,
        ),
        _get_statistics(
            config=config,
            name="Providers",
            description="Search providers, i.e., websites that offer a search functionality.",
            document=Provider,
            index=config.es.index_providers,
        ),
        _get_statistics(
            config=config,
            name="Sources",
            description="The cross product of all archives and the provider's domains and URL prefixes.",
            document=Source,
            index=config.es.index_sources,
        ),
        _get_statistics(
            config=config,
            name="Captures",
            description="Captures matching from the archives that match domain and URL prefixes.",
            document=Capture,
            index=config.es.index_captures,
        ),
        _get_statistics(
            config=config,
            name="SERPs",
            description="Search engine result pages that have been identified among the captures.",
            document=Serp,
            index=config.es.index_serps,
        ),
        _get_statistics(
            config=config,
            name="+ URL query",
            description="SERPs for which the query has been parsed from the URL.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="url_query"),
        ),
        _get_statistics(
            config=config,
            name="+ URL page",
            description="SERPs for which the page has been parsed from the URL.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="url_page"),
        ),
        _get_statistics(
            config=config,
            name="+ URL offset",
            description="SERPs for which the offset has been parsed from the URL.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="url_offset"),
        ),
        _get_statistics(
            config=config,
            name="+ WARC",
            description="SERPs for which the WARC has been downloaded.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="warc_location"),
        ),
        _get_statistics(
            config=config,
            name="+ WARC query",
            description="SERPs for which the query has been parsed from the WARC.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="warc_query"),
        ),
        _get_statistics(
            config=config,
            name="+ WARC snippets",
            description="SERPs for which the snippets have been parsed from the WARC.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="warc_snippets_parser.id"),
        ),
        _get_statistics(
            config=config,
            name="Results",
            description="Search result from the SERPs.",
            document=Result,
            index=config.es.index_results,
        ),
        _get_statistics(
            config=config,
            name="URL query parsers",
            description="Parser to get the query from a SERP's URL.",
            document=UrlQueryParser,
            index=config.es.index_url_query_parsers,
        ),
        _get_statistics(
            config=config,
            name="URL page parsers",
            description="Parser to get the page from a SERP's URL.",
            document=UrlPageParser,
            index=config.es.index_url_page_parsers,
        ),
        _get_statistics(
            config=config,
            name="URL offset parsers",
            description="Parser to get the offset from a SERP's URL.",
            document=UrlOffsetParser,
            index=config.es.index_url_offset_parsers,
        ),
        _get_statistics(
            config=config,
            name="WARC query parsers",
            description="Parser to get the query from a SERP's WARC contents.",
            document=WarcQueryParser,
            index=config.es.index_warc_query_parsers,
        ),
        _get_statistics(
            config=config,
            name="WARC snippets parsers",
            description="Parser to get the snippets from a SERP's WARC contents.",
            document=WarcSnippetsParser,
            index=config.es.index_warc_snippets_parsers,
        ),
    ]

    return statistics_list


@app.get("/progress")
def get_progress() -> list[Progress]:
    config: Config = Config()

    progress_list = [
        _get_processed_progress(
            config=config,
            input_name="Archives",
            output_name="Sources",
            description="Build sources for all archives.",
            document=Archive,
            index=config.es.index_archives,
            filter_query=~Exists(field="exclusion_reason"),
            status_field="should_build_sources",
        ),
        _get_processed_progress(
            config=config,
            input_name="Providers",
            output_name="Sources",
            description="Build sources for all search providers.",
            document=Provider,
            index=config.es.index_providers,
            filter_query=~Exists(field="exclusion_reason"),
            status_field="should_build_sources",
        ),
        _get_processed_progress(
            config=config,
            input_name="Sources",
            output_name="Captures",
            description="Fetch CDX captures for all domains and prefixes in the sources.",
            document=Source,
            index=config.es.index_sources,
            status_field="should_fetch_captures",
        ),
        _get_processed_progress(
            config=config,
            input_name="Captures",
            output_name="SERPs",
            description="Parse queries from capture URLs.",
            document=Capture,
            index=config.es.index_captures,
            status_field="url_query_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse page from SERP URLs.",
            document=Serp,
            index=config.es.index_serps,
            status_field="url_page_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse offset from SERP URLs.",
            document=Serp,
            index=config.es.index_serps,
            status_field="url_offset_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Download WARCs.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Term(capture__status_code=200),
            status_field="warc_downloader.should_download",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse query from WARC contents.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="warc_location"),
            status_field="warc_query_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse snippets from WARC contents.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="warc_location"),
            status_field="warc_snippets_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="Results",
            output_name="Results",
            description="Download WARCs.",
            document=Result,
            index=config.es.index_results,
            filter_query=Exists(field="snippet.url"),
            status_field="warc_downloader.should_download",
        ),
    ]

    return progress_list



from pathlib import Path

from click import group, option, Path as PathType, FloatRange

from archive_query_log.cli.util import validate_split_domains, pass_config
from archive_query_log.config import Config
from archive_query_log.orm import Provider


@group()
def providers() -> None:
    pass


@providers.command()
@option("--name", type=str)
@option("--description", type=str)
@option("--notes", type=str)
@option("--exclusion-reason", "--exclusion", type=str)
@option("--domains", "--domain", type=str, multiple=True,
        required=True, callback=validate_split_domains)
@option("--url-path-prefixes", "--url-path-prefix", type=str,
        multiple=True, required=True, metavar="PREFIXES")
@option("--priority", type=FloatRange(min=0, min_open=False))
@pass_config
def add(
        config: Config,
        name: str | None,
        description: str | None,
        notes: str | None,
        exclusion_reason: str | None,
        domains: list[str],
        url_path_prefixes: list[str],
        priority: float | None,
) -> None:
    from archive_query_log.providers import add_provider
    Provider.init(using=config.es.client, index=config.es.index_providers)
    add_provider(
        config=config,
        name=name,
        description=description,
        notes=notes,
        exclusion_reason=exclusion_reason,
        domains=set(domains),
        url_path_prefixes=set(url_path_prefixes),
        priority=priority,
    )


@providers.command("import")
@option("-s", "--services-file", "services_path",
        type=PathType(path_type=Path, exists=True, file_okay=True,
                      dir_okay=False, readable=True, resolve_path=True,
                      allow_dash=False),
        default=Path("data") / "selected-services.yaml")
@option("-c", "--cache-dir", "cache_path",
        type=PathType(path_type=Path, exists=False, file_okay=False,
                      dir_okay=True, readable=True, writable=True,
                      resolve_path=True, allow_dash=False),
        default=Path("data") / "cache" / "provider-names")
@option("--review", type=int)
@option("--no-merge", is_flag=True, default=False, type=bool)
@option("--auto-merge", is_flag=True, default=False, type=bool)
@pass_config
def import_(
        config: Config,
        services_path: Path,
        cache_path: Path,
        review: int | None,
        no_merge: bool,
        auto_merge: bool,
) -> None:
    from archive_query_log.imports.yaml import import_providers
    Provider.init(using=config.es.client, index=config.es.index_providers)
    import_providers(
        config=config,
        services_path=services_path,
        cache_path=cache_path,
        review=review,
        no_merge=no_merge,
        auto_merge=auto_merge,
    )



from dataclasses import dataclass
from gzip import GzipFile
from pathlib import Path
from typing import Sequence, NamedTuple, Pattern, Iterable
from urllib.parse import parse_qsl, unquote, quote

from tqdm.auto import tqdm

from archive_query_log.legacy.model import ArchivedQueryUrl, \
    ArchivedUrl, PageParser, QueryParser, OffsetParser, Service
from archive_query_log.legacy.urls.iterable import ArchivedUrls
from archive_query_log.legacy.util.text import text_io_wrapper


@dataclass(frozen=True)
class QueryParameterQueryParser(QueryParser):
    url_pattern: Pattern[str]
    parameter: str

    def parse(self, url: ArchivedUrl) -> str | None:
        if self.url_pattern.search(url.url) is None:
            return None
        for key, value in parse_qsl(url.split_url.query):
            if key == self.parameter:
                return value.strip()
        return None


@dataclass(frozen=True)
class FragmentParameterQueryParser(QueryParser):
    url_pattern: Pattern[str]
    parameter: str

    def parse(self, url: ArchivedUrl) -> str | None:
        if self.url_pattern.search(url.url) is None:
            return None
        for key, value in parse_qsl(url.split_url.fragment):
            if key == self.parameter:
                return value.strip()
        return None


@dataclass(frozen=True)
class PathSegmentQueryParser(QueryParser):
    url_pattern: Pattern[str]
    segment: int
    delimiter: str = "/"
    remove_patterns: Sequence[Pattern[str]] = tuple()
    space_patterns: Sequence[Pattern[str]] = tuple()

    def parse(self, url: ArchivedUrl) -> str | None:
        if self.url_pattern.search(url.url) is None:
            return None
        path = url.split_url.path
        path_segments = path.split(self.delimiter)
        if len(path_segments) <= self.segment:
            return None
        path_segment = path_segments[self.segment]
        for pattern in self.remove_patterns:
            path_segment = pattern.sub("", path_segment)
        if len(self.space_patterns) > 0:
            for pattern in self.space_patterns:
                path_segment = pattern.sub(" ", path_segment)
            path_segment = path_segment.replace("  ", " ")
        return unquote(path_segment).strip()


@dataclass(frozen=True)
class FragmentSegmentQueryParser(QueryParser):
    url_pattern: Pattern[str]
    segment: int
    delimiter: str = "/"
    remove_patterns: Sequence[Pattern[str]] = tuple()
    space_patterns: Sequence[Pattern[str]] = tuple()

    def parse(self, url: ArchivedUrl) -> str | None:
        if self.url_pattern.search(url.url) is None:
            return None
        path = url.split_url.fragment
        path_segments = path.split(self.delimiter)
        if len(path_segments) <= self.segment:
            return None
        path_segment = path_segments[self.segment]
        for pattern in self.remove_patterns:
            path_segment = pattern.sub("", path_segment)
        if len(self.space_patterns) > 0:
            for pattern in self.space_patterns:
                path_segment = pattern.sub(" ", path_segment)
            path_segment = path_segment.replace("  ", " ")
        return unquote(path_segment).strip()


@dataclass(frozen=True)
class QueryParameterPageOffsetParser(PageParser, OffsetParser):
    url_pattern: Pattern[str]
    parameter: str

    def parse(self, url: ArchivedUrl) -> int | None:
        if self.url_pattern.search(url.url) is None:
            return None
        for key, value in parse_qsl(url.split_url.query):
            if key == self.parameter and value.isdigit():
                return int(value)
        return None


@dataclass(frozen=True)
class FragmentParameterPageOffsetParser(PageParser, OffsetParser):
    url_pattern: Pattern[str]
    parameter: str

    def parse(self, url: ArchivedUrl) -> int | None:
        if self.url_pattern.search(url.url) is None:
            return None
        for key, value in parse_qsl(url.split_url.fragment):
            if key == self.parameter and value.isdigit():
                return int(value)
        return None


@dataclass(frozen=True)
class PathSegmentPageOffsetParser(PageParser, OffsetParser):
    url_pattern: Pattern[str]
    segment: int
    delimiter: str = "/"
    remove_patterns: Sequence[Pattern[str]] = tuple()

    def parse(self, url: ArchivedUrl) -> int | None:
        if self.url_pattern.search(url.url) is None:
            return None
        path = url.split_url.path
        path_segments = path.split(self.delimiter)
        if len(path_segments) <= self.segment:
            return None
        path_segment = path_segments[self.segment]
        for pattern in self.remove_patterns:
            path_segment = pattern.sub("", path_segment)
        return int(path_segment)


@dataclass(frozen=True)
class FragmentSegmentPageOffsetParser(PageParser, OffsetParser):
    url_pattern: Pattern[str]
    segment: int
    delimiter: str = "/"
    remove_patterns: Sequence[Pattern[str]] = tuple()

    def parse(self, url: ArchivedUrl) -> int | None:
        if self.url_pattern.search(url.url) is None:
            return None
        path = url.split_url.fragment
        path_segments = path.split(self.delimiter)
        if len(path_segments) <= self.segment:
            return None
        path_segment = path_segments[self.segment]
        for pattern in self.remove_patterns:
            path_segment = pattern.sub("", path_segment)
        return int(path_segment)


class _CdxPage(NamedTuple):
    input_path: Path
    output_path: Path


@dataclass(frozen=True)
class ArchivedQueryUrlParser:
    query_parsers: Sequence[QueryParser]
    page_parsers: Sequence[PageParser]
    offset_parsers: Sequence[OffsetParser]
    overwrite: bool = False
    verbose: bool = False

    def parse(
            self,
            input_path: Path,
            output_path: Path,
            focused: bool = False,
    ) -> None:
        if output_path.exists() and not self.overwrite:
            return
        output_path.parent.mkdir(parents=True, exist_ok=True)
        archived_urls: Iterable[ArchivedUrl] = ArchivedUrls(input_path)
        if self.verbose:
            # noinspection PyTypeChecker
            archived_urls = tqdm(
                archived_urls,
                desc="Parse SERP URLs",
                unit="URL",
            )
        archived_serp_urls_nullable = (
            self._parse_single(archived_url, focused)
            for archived_url in archived_urls
        )
        archived_serp_urls = (
            archived_serp_url
            for archived_serp_url in archived_serp_urls_nullable
            if archived_serp_url is not None
        )
        output_schema = ArchivedQueryUrl.schema()
        with (output_path.open("wb") as file,
              GzipFile(fileobj=file, mode="wb") as gzip_file,
              text_io_wrapper(gzip_file) as text_file):
            for archived_serp_url in archived_serp_urls:
                text_file.write(output_schema.dumps(archived_serp_url))
                text_file.write("\n")

    def _parse_single(
            self,
            archived_url: ArchivedUrl,
            focused: bool,
    ) -> ArchivedQueryUrl | None:
        query: str | None = None
        for query_parser in self.query_parsers:
            query = query_parser.parse(archived_url)
            if query is not None:
                break

        if query is None:
            return None

        page: int | None = None
        for page_parser in self.page_parsers:
            page = page_parser.parse(archived_url)
            if page is not None:
                break

        if focused and page is not None and page != 0:
            return None

        offset: int | None = None
        for offset_parser in self.offset_parsers:
            offset = offset_parser.parse(archived_url)
            if offset is not None:
                break

        if focused and offset is not None and offset != 0:
            return None

        return ArchivedQueryUrl(
            url=archived_url.url,
            timestamp=archived_url.timestamp,
            query=query,
            page=page,
            offset=offset,
        )

    def _service_pages(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None,
            cdx_page: int | None,
    ) -> Sequence[_CdxPage]:
        """
        List all items that need to be downloaded.
        """
        input_format_path = data_directory / "archived-urls"
        output_format_path = data_directory / "archived-query-urls"

        service_path = input_format_path / service.name

        if domain is not None:
            domain_paths = [service_path / domain]
        else:
            domain_paths = [
                path
                for path in service_path.iterdir()
                if path.is_dir()
            ]
            if focused:
                domain_paths = [
                    path
                    for path in domain_paths
                    if any(
                        path.name.endswith(quote(prefix, safe=""))
                        for prefix in service.focused_url_prefixes
                    )
                ]

        if cdx_page is not None:
            if domain is None:
                raise RuntimeError(
                    "Domain must be specified when page is specified.")
            if len(domain_paths) < 1:
                raise RuntimeError(
                    "There must be exactly one domain path.")
            cdx_page_paths = [domain_paths[0] / f"{cdx_page:010}.jsonl.gz"]
        else:
            cdx_page_paths = [
                path
                for domain_path in domain_paths
                for path in domain_path.iterdir()
                if (
                        path.is_file() and
                        len(path.name.removesuffix(".jsonl.gz")) == 10 and
                        path.name.removesuffix(".jsonl.gz").isdigit()
                )
            ]

        return [
            _CdxPage(
                input_path=cdx_page_path,
                output_path=output_format_path / cdx_page_path.relative_to(
                    input_format_path
                ),
            )
            for cdx_page_path in cdx_page_paths
        ]

    def parse_service(
            self,
            data_directory: Path,
            focused: bool,
            service: Service,
            domain: str | None = None,
            cdx_page: int | None = None,
    ):
        pages_list: Sequence[_CdxPage] = self._service_pages(
            data_directory=data_directory,
            focused=focused,
            service=service,
            domain=domain,
            cdx_page=cdx_page,
        )

        if len(pages_list) == 0:
            return

        pages: Iterable[_CdxPage] = pages_list
        if len(pages_list) > 1:
            # noinspection PyTypeChecker
            pages = tqdm(
                pages,
                desc="Parse archived SERP URLs",
                unit="page",
            )

        for page in pages:
            self.parse(page.input_path, page.output_path, focused)



from archive_query_log.legacy.config import SERVICES

if __name__ == '__main__':
    num_url_prefixes = sum(
        len(service.focused_url_prefixes)
        for service in SERVICES.values()
    )
    print(f"Number of URL prefixes: {num_url_prefixes}")
    num_query_parsers = sum(
        len(service.query_parsers)
        for service in SERVICES.values()
    )
    print(f"Number of query parsers: {num_query_parsers}")
    num_page_parsers = sum(
        len(service.page_parsers)
        for service in SERVICES.values()
    )
    print(f"Number of page parsers: {num_page_parsers}")
    num_offset_parsers = sum(
        len(service.offset_parsers)
        for service in SERVICES.values()
    )
    print(f"Number of offset parsers: {num_offset_parsers}")
    num_interpreted_query_parsers = sum(
        len(service.interpreted_query_parsers)
        for service in SERVICES.values()
    )
    print(
        f"Number of interpreted query parsers: {num_interpreted_query_parsers}"
    )
    num_results_parsers = sum(
        len(service.results_parsers)
        for service in SERVICES.values()
    )
    print(f"Number of results parsers: {num_results_parsers}")



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_weibo_sheng_fu_lang_xi_si_1603546157():
    verify_serp_parsing(
        "https://web.archive.org/web/20201024132917id_/https://s.weibo.com/weibo?q=%23%E5%9C%A3%C2%B7%E5%BC%97%E6%9C%97%E8%A5%BF%E6%96%AF%23&xsort=hot&Refer=hotmore",
        "weibo",
    )


def test_parse_query_weibo_bu_xiang_jiao_hun_li_de_fen_zi_qian_refer_focus_lx_stopic_box_1527256388():
    verify_serp_parsing(
        "https://web.archive.org/web/20180525135308id_/http://s.weibo.com:80/weibo/%E4%B8%8D%E6%83%B3%E4%BA%A4%E5%A9%9A%E7%A4%BC%E7%9A%84%E4%BB%BD%E5%AD%90%E9%92%B1&Refer=focus_lx_STopic_box",
        "weibo",
    )


def test_parse_query_weibo_ji_huo_ma_b_1_page_2_1433297692():
    verify_serp_parsing(
        "https://web.archive.org/web/20150603021452id_/http://s.weibo.com:80/weibo/%E6%BF%80%E6%B4%BB%E7%A0%81&b=1&page=2",
        "weibo",
    )


def test_parse_query_weibo_li_wei_yi_refer_user_weibo_1517733046():
    verify_serp_parsing(
        "https://web.archive.org/web/20180204083046id_/http://s.weibo.com:80/weibo/%E6%9D%8E%E5%BE%AE%E6%BC%AA&Refer=user_weibo",
        "weibo",
    )


def test_parse_query_weibo_hun_li_li_jie_refer_stopic_box_1505917853():
    verify_serp_parsing(
        "https://web.archive.org/web/20170920143053id_/http://s.weibo.com:80/weibo/%E5%A9%9A%E7%A4%BC%E7%A4%BC%E8%8A%82&Refer=STopic_box",
        "weibo",
    )


def test_parse_query_weibo_kan_bu_jian_de_yan_pi_1409920649():
    verify_serp_parsing(
        "https://web.archive.org/web/20140905123729id_/http://s.weibo.com:80/weibo/%E7%9C%8B%E4%B8%8D%E8%A7%81%E5%BE%97%E7%9C%BC%E7%9A%AE?",
        "weibo",
    )


def test_parse_query_weibo_xiao_xiao_bin_refer_stopic_box_1518290095():
    verify_serp_parsing(
        "https://web.archive.org/web/20180210191455id_/http://s.weibo.com:80/weibo/%E5%B0%8F%E5%B0%8F%E5%BD%AC&Refer=STopic_box",
        "weibo",
    )


def test_parse_query_weibo_meng_lin_dexiao_wu_1437753277():
    verify_serp_parsing(
        "https://web.archive.org/web/20150724155437id_/http://s.weibo.com:80/weibo/%E5%AD%9F%E6%9E%97de%E5%B0%8F%E5%B1%8B?",
        "weibo",
    )


def test_parse_query_weibo_fu_cai_3dshu_ju_jia_wei_xin_xw639_2018_1555261848():
    verify_serp_parsing(
        "https://web.archive.org/web/20190414171048id_/https://s.weibo.com/weibo/%E7%A6%8F%E5%BD%A93d%E6%95%B0%E6%8D%AE%E5%8A%A0%E5%BE%AE%E2%98%85%E4%BF%A1xw639-2018%E2%80%BB",
        "weibo",
    )


def test_parse_query_weibo_quan_ye_cha_page_13_1517932436():
    verify_serp_parsing(
        "https://web.archive.org/web/20180206155356id_/http://s.weibo.com:80/weibo/%E7%8A%AC%E5%A4%9C%E5%8F%89&page=13",
        "weibo",
    )



from collections import defaultdict
from datetime import datetime, timezone
from gzip import GzipFile
from io import TextIOWrapper, BytesIO
from json import loads
from pathlib import Path
from random import Random
from re import compile as pattern
from textwrap import dedent

from requests import get
from slugify import slugify
from tqdm import tqdm
from warcio import WARCWriter, StatusAndHeaders

from archive_query_log.legacy import PROJECT_DIRECTORY_PATH
from archive_query_log.legacy.model import ArchivedQueryUrl

NUM_SERVICES = 11
SERVICE_NAMES = ["google", "yahoo", "bing", "duckduckgo", "ask", "ecosia"]
NUM_QUERIES_PER_SERVICE = 50

DATA_PATH = Path(
    "/mnt/ceph/storage/data-in-progress/data-research/"
    "web-search/web-archive-query-log/focused/"
)

SAMPLE_QUERIES_PATH = DATA_PATH / "corpus" / "medium" / "2023-05-22" / "serps"

WARCS_PATH = PROJECT_DIRECTORY_PATH / \
             "data/manual-annotations/archived-raw-serps/warcs/"
TESTS_PATH = PROJECT_DIRECTORY_PATH / \
             "archive_query_log/results/test/"

PATTERN_SPECIAL_CHARS = pattern(r"[^0-9a-z]+")


def warc_url(url: str, timestamp: float) -> str:
    wayback_timestamp = datetime \
        .fromtimestamp(timestamp, timezone.utc) \
        .strftime("%Y%m%d%H%M%S")
    wayback_raw_url = \
        f"https://web.archive.org/web/{wayback_timestamp}id_/{url}"
    return wayback_raw_url


def main():
    service_names = SERVICE_NAMES

    query_urls = defaultdict(list)
    for path in tqdm(
            list(SAMPLE_QUERIES_PATH.glob("part*.gz")),
            desc="Load service queries"
    ):
        # noinspection PyTypeChecker
        with GzipFile(path, "rb") as gf, TextIOWrapper(gf) as f:
            for line in f:
                if "\"serp_query_text_url\": \"" not in line:
                    continue
                if "\"serp_warc_relative_path\": \"" not in line:
                    continue
                query_url = loads(line)
                if query_url["search_provider_name"] not in service_names:
                    continue
                query_urls[query_url["search_provider_name"]].append(query_url)

    print(f"Found {sum(len(urls) for urls in query_urls.values())} SERPs.")
    random = Random(0)  # nosec: B311
    query_urls = {
        service_name: random.sample(
            query_urls[service_name], min(
                NUM_QUERIES_PER_SERVICE,
                len(query_urls[service_name]),
            )
        )
        for service_name in service_names
    }
    print(f"Sampled {sum(len(urls) for urls in query_urls.values())} SERPs.")

    print("Generate WARC files.")
    schema = ArchivedQueryUrl.schema()
    for service_name in service_names:
        for query_url in tqdm(
                query_urls[service_name], desc=service_name
        ):
            query = query_url["serp_query_text_url"]
            query = slugify(query)
            query = query[:100]
            name = slugify(
                f"{service_name}-"
                f"{query}-{query_url['serp_timestamp']}"
            )
            warc_path = WARCS_PATH / f"{name}.warc.gz"
            if warc_path.exists():
                continue
            with warc_path.open("wb") as o:
                writer = WARCWriter(o, gzip=True)
                archived_query_url = ArchivedQueryUrl(
                    url=query_url["serp_url"],
                    timestamp=int(query_url["serp_timestamp"]),
                    query=query_url["serp_query_text_url"],
                    page=query_url["serp_page"],
                    offset=query_url["serp_offset"],
                )
                url_headers = {
                    "Archived-URL": schema.dumps(archived_query_url),
                }
                wayback_raw_url = warc_url(
                    query_url["serp_url"],
                    int(query_url["serp_timestamp"]),
                )
                response = get(
                    wayback_raw_url,
                    timeout=60 * 4,  # nosec: B113
                )
                response.raise_for_status()

                # noinspection PyProtectedMember
                version = str(response.raw.version)
                protocol = f"HTTP/{version[0]}.{version[1]}"
                request_record = writer.create_warc_record(
                    uri=str(response.request.url),
                    record_type="request",
                    http_headers=StatusAndHeaders(
                        statusline=" ".join((
                            response.request.method,
                            response.request.path_url,
                            protocol,
                        )),
                        headers=response.request.headers,
                        protocol=protocol,
                    ),
                    warc_headers_dict={**url_headers},
                )
                writer.write_record(request_record)
                response_record = writer.create_warc_record(
                    uri=str(response.url),
                    record_type="response",
                    http_headers=StatusAndHeaders(
                        statusline=" ".join((
                            protocol,
                            str(response.status_code),
                            response.reason,
                        )),
                        headers=response.headers,
                        protocol=protocol
                    ),
                    payload=BytesIO(response.content),
                    length=len(response.content),
                    warc_headers_dict={**url_headers},
                )
                writer.write_record(response_record)

    print("Generate test files.")
    for service_name in service_names:
        if len(query_urls[service_name]) == 0:
            continue
        test_path = TESTS_PATH / f"test_{service_name}_serp_parsing.py"
        if not test_path.exists():
            with test_path.open("wt") as o:
                o.write(dedent("""
                # flake8: noqa
                # This file is auto-generated by generate_tests.py.
                """).lstrip())
                o.write(dedent("""
                from archive_query_log.results.test.test_utils import verify_serp_parsing
                """).lstrip())
        with test_path.open("rt") as f:
            existing_tests = {
                line.removeprefix("def test_parse_query_").removesuffix("():")
                for line in f
                if line.startswith("def test_parse_query_")
            }
        with test_path.open("at") as o:
            for query_url in query_urls[service_name]:
                wayback_raw_url = warc_url(
                    query_url["serp_url"],
                    int(query_url["serp_timestamp"]),
                )

                query = query_url["serp_query_text_url"]
                query = slugify(query)
                query = query[:100]
                name = slugify(
                    f"{service_name}_{query}_{query_url['serp_timestamp']}",
                    separator="_",
                )
                wayback_raw_url_safe = wayback_raw_url.replace('"', '\\"')

                if name in existing_tests:
                    continue
                o.write(dedent(f"""

                def test_parse_query_{name}():
                    verify_serp_parsing(
                        "{wayback_raw_url_safe}",
                        "{service_name}",
                    )
                """))


if __name__ == '__main__':
    main()



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_youtube_pudding_1563068696():
    verify_serp_parsing(
        "https://web.archive.org/web/20190714014456id_/https://www.youtube.com/results?search_query=%23pudding&sp=EgQYASABSGTqAwA%253D",
        "youtube",
    )


def test_parse_query_youtube_q2_2017_arizona_west_regional_1547871680():
    verify_serp_parsing(
        "https://web.archive.org/web/20190119042120id_/https://www.youtube.com/results?search_query=Q2+2017+Arizona%20West%20Regional",
        "youtube",
    )


def test_parse_query_youtube_lataji_1563086980():
    verify_serp_parsing(
        "https://web.archive.org/web/20190714064940id_/https://www.youtube.com/results?search_query=%23LataJi&sp=CAESBhABIAEoAQ%253D%253D",
        "youtube",
    )


def test_parse_query_youtube_robloks_mip_siti_2019_1583715307():
    verify_serp_parsing(
        "https://web.archive.org/web/20200309005507id_/https://www.youtube.com/results?search_query=%D1%80%D0%BE%D0%B1%D0%BB%D0%BE%D0%BA%D1%81+%D0%BC%D0%B8%D0%BF+%D1%81%D0%B8%D1%82%D0%B8+2019&sp=SDzqAwA%253D",
        "youtube",
    )


def test_parse_query_youtube_ampatuanmassacre_1583309425():
    verify_serp_parsing(
        "https://web.archive.org/web/20200304081025id_/https://www.youtube.com/results?search_query=%23AmpatuanMassacre&app=desktop",
        "youtube",
    )


def test_parse_query_youtube_hp_probook_640_g1_razborka_1579996098():
    verify_serp_parsing(
        "https://web.archive.org/web/20200125234818id_/https://www.youtube.com/results?search_query=hp+probook+640+g1+%D1%80%D0%B0%D0%B7%D0%B1%D0%BE%D1%80%D0%BA%D0%B0",
        "youtube",
    )


def test_parse_query_youtube_razbor_shchetki_pylesosa_samsung_1584184489():
    verify_serp_parsing(
        "https://web.archive.org/web/20200314111449id_/https://www.youtube.com/results?search_query=%D1%80%D0%B0%D0%B7%D0%B1%D0%BE%D1%80+%D1%89%D0%B5%D1%82%D0%BA%D0%B8+%D0%BF%D1%8B%D0%BB%D0%B5%D1%81%D0%BE%D1%81%D0%B0+samsung",
        "youtube",
    )


def test_parse_query_youtube_literatura_6_klass_biografiia_pushkina_1595705363():
    verify_serp_parsing(
        "https://web.archive.org/web/20200725192923id_/https://www.youtube.com/results?search_query=%D0%BB%D0%B8%D1%82%D0%B5%D1%80%D0%B0%D1%82%D1%83%D1%80%D0%B0+6+%D0%BA%D0%BB%D0%B0%D1%81%D1%81+%D0%B1%D0%B8%D0%BE%D0%B3%D1%80%D0%B0%D1%84%D0%B8%D1%8F+%D0%BF%D1%83%D1%88%D0%BA%D0%B8%D0%BD%D0%B0",
        "youtube",
    )


def test_parse_query_youtube_ne_prosto_bekkhen_reaktsiia_1582514079():
    verify_serp_parsing(
        "https://web.archive.org/web/20200224031439id_/https://www.youtube.com/results?search_query=%D0%BD%D0%B5+%D0%BF%D1%80%D0%BE%D1%81%D1%82%D0%BE+%D0%B1%D1%8D%D0%BA%D1%85%D0%B5%D0%BD+%D1%80%D0%B5%D0%B0%D0%BA%D1%86%D0%B8%D1%8F",
        "youtube",
    )


def test_parse_query_youtube_prokhozhdenie_igry_madagaskar_2_chast_4_1562167916():
    verify_serp_parsing(
        "https://web.archive.org/web/20190703153156id_/https://www.youtube.com/results?search_query=%D0%BF%D1%80%D0%BE%D1%85%D0%BE%D0%B6%D0%B4%D0%B5%D0%BD%D0%B8%D0%B5+%D0%B8%D0%B3%D1%80%D1%8B+%D0%BC%D0%B0%D0%B4%D0%B0%D0%B3%D0%B0%D1%81%D0%BA%D0%B0%D1%80+2+%D1%87%D0%B0%D1%81%D1%82%D1%8C+4&sp=CAESCwgFEAQYAiAByAEB",
        "youtube",
    )


def test_parse_query_youtube_zadnii_most_moskvich_412_ustroistvo_1584150216():
    verify_serp_parsing(
        "https://web.archive.org/web/20200314014336id_/https://www.youtube.com/results?search_query=%D0%B7%D0%B0%D0%B4%D0%BD%D0%B8%D0%B9+%D0%BC%D0%BE%D1%81%D1%82+%D0%BC%D0%BE%D1%81%D0%BA%D0%B2%D0%B8%D1%87+412+%D1%83%D1%81%D1%82%D1%80%D0%BE%D0%B9%D1%81%D1%82%D0%B2%D0%BE",
        "youtube",
    )


def test_parse_query_youtube_vasilii_emelianenko_tefteli_1578285143():
    verify_serp_parsing(
        "https://web.archive.org/web/20200106043223id_/https://www.youtube.com/results?search_query=%D0%B2%D0%B0%D1%81%D0%B8%D0%BB%D0%B8%D0%B9+%D0%B5%D0%BC%D0%B5%D0%BB%D1%8C%D1%8F%D0%BD%D0%B5%D0%BD%D0%BA%D0%BE+%D1%82%D0%B5%D1%84%D1%82%D0%B5%D0%BB%D0%B8",
        "youtube",
    )


def test_parse_query_youtube_kutyaplya_1561171748():
    verify_serp_parsing(
        "https://web.archive.org/web/20190622024908id_/https://www.youtube.com/results?search_query=kutyaplya",
        "youtube",
    )


def test_parse_query_youtube_kak_sdelat_vzryv_v_cinema_4d_1578732009():
    verify_serp_parsing(
        "https://web.archive.org/web/20200111084009id_/https://www.youtube.com/results?search_query=%D0%BA%D0%B0%D0%BA+%D1%81%D0%B4%D0%B5%D0%BB%D0%B0%D1%82%D1%8C+%D0%B2%D0%B7%D1%80%D1%8B%D0%B2+%D0%B2+cinema+4d",
        "youtube",
    )


def test_parse_query_youtube_shang_yue_xian_rupu_1582390054():
    verify_serp_parsing(
        "https://web.archive.org/web/20200222164734id_/https://www.youtube.com/results?search_query=%E4%B8%8A%E8%B6%8A%E7%B7%9A+%E3%83%AB%E3%83%BC%E3%83%97",
        "youtube",
    )


def test_parse_query_youtube_pora_tiulpanov_aktery_1574991226():
    verify_serp_parsing(
        "https://web.archive.org/web/20191129013346id_/https://www.youtube.com/results?search_query=%D0%BF%D0%BE%D1%80%D0%B0+%D1%82%D1%8E%D0%BB%D1%8C%D0%BF%D0%B0%D0%BD%D0%BE%D0%B2+%D0%B0%D0%BA%D1%82%D0%B5%D1%80%D1%8B",
        "youtube",
    )


def test_parse_query_youtube_uzbekskaia_svadba_v_shymkente_1574145605():
    verify_serp_parsing(
        "https://web.archive.org/web/20191119064005id_/https://www.youtube.com/results?search_query=%D1%83%D0%B7%D0%B1%D0%B5%D0%BA%D1%81%D0%BA%D0%B0%D1%8F+%D1%81%D0%B2%D0%B0%D0%B4%D1%8C%D0%B1%D0%B0+%D0%B2+%D1%88%D1%8B%D0%BC%D0%BA%D0%B5%D0%BD%D1%82%D0%B5",
        "youtube",
    )


def test_parse_query_youtube_chto_budet_esli_sobrat_vse_rezinki_v_bully_1599555287():
    verify_serp_parsing(
        "https://web.archive.org/web/20200908085447id_/https://www.youtube.com/results?search_query=%D1%87%D1%82%D0%BE+%D0%B1%D1%83%D0%B4%D0%B5%D1%82+%D0%B5%D1%81%D0%BB%D0%B8+%D1%81%D0%BE%D0%B1%D1%80%D0%B0%D1%82%D1%8C+%D0%B2%D1%81%D0%B5+%D1%80%D0%B5%D0%B7%D0%B8%D0%BD%D0%BA%D0%B8+%D0%B2+bully",
        "youtube",
    )


def test_parse_query_youtube_kak_vybit_dushu_moba_1599057231():
    verify_serp_parsing(
        "https://web.archive.org/web/20200902143351id_/https://www.youtube.com/results?search_query=%D0%BA%D0%B0%D0%BA+%D0%B2%D1%8B%D0%B1%D0%B8%D1%82%D1%8C+%D0%B4%D1%83%D1%88%D1%83+%D0%BC%D0%BE%D0%B1%D0%B0",
        "youtube",
    )


def test_parse_query_youtube_cardistry_asmr_1577113546():
    verify_serp_parsing(
        "https://web.archive.org/web/20191223150546id_/https://www.youtube.com/results?search_query=cardistry+asmr",
        "youtube",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_yandex_speed_force_1535973684():
    verify_serp_parsing(
        "https://web.archive.org/web/20180903112124id_/https://yandex.ru/images/search?p=21&text=speed%20force&img_url=https%3A%2F%2Fi.ytimg.com%2Fvi%2Fy2J5Bq71emM%2Fmqdefault.jpg&pos=648&rpt=simage",
        "yandex",
    )


def test_parse_query_yandex_speed_force_1535970436():
    verify_serp_parsing(
        "https://web.archive.org/web/20180903102716id_/https://yandex.ru/images/search?p=12&text=speed%20force&img_url=https%3A%2F%2Fmedia.dayoftheshirt.com%2Fimages%2Fshirts%2F2rR4SxCzB2Yv%2Fonceuponatee_speed-force-ii_1459188765.full.png&pos=374&rpt=simage",
        "yandex",
    )


def test_parse_query_yandex_speed_force_1536100368():
    verify_serp_parsing(
        "https://web.archive.org/web/20180904223248id_/https://yandex.ru/images/search?p=16&text=speed%20force&img_url=http%3A%2F%2Fs3.birthmoviesdeath.com%2Fimages%2Fmade%2FFFSF4_1200_1220_81_s.jpg&pos=502&rpt=simage",
        "yandex",
    )


def test_parse_query_yandex_spirited_away_animated_film_2001_1524478207():
    verify_serp_parsing(
        "https://web.archive.org/web/20180423101007id_/https://yandex.ru/video/search?text=spirited%20away%20animated%20film%202001&source=recommendations&path=main&personal=false&autoplay=1",
        "yandex",
    )


def test_parse_query_yandex_speed_force_1536042339():
    verify_serp_parsing(
        "https://web.archive.org/web/20180904062539id_/https://yandex.ru/images/search?p=2&text=speed%20force&img_url=https%3A%2F%2Fandroidapplications.ru%2Fuploads%2Fposts%2F2017-04%2F1492531591_speed_force_race_-_gonki-6.jpg&pos=82&rpt=simage",
        "yandex",
    )


def test_parse_query_yandex_danses_tv_periodic_1534841786():
    verify_serp_parsing(
        "https://web.archive.org/web/20180821085626id_/https://yandex.ru/video/search?text=danses%20tv%20periodic&source=recommendations&path=main&within=9&personal=false",
        "yandex",
    )


def test_parse_query_yandex_virat_swaroop_1523992633():
    verify_serp_parsing(
        "https://web.archive.org/web/20180417191713id_/https://yandex.ru/images/search?text=virat%20swaroop",
        "yandex",
    )


def test_parse_query_yandex_niusha_1515577404():
    verify_serp_parsing(
        "https://web.archive.org/web/20180110094324id_/https://yandex.ru/video/search?text=%D0%9D%D1%8E%D1%88%D0%B0&source=oo&entref=0oCgpydXcyNDk5MjQzEg9ydXczOTUwNDctYXNzb2MYAt5s7-k&parent-reqid=1515552693812954-1072433413594552848811520-vla1-2341-V",
        "yandex",
    )


def test_parse_query_yandex_speed_force_1535955246():
    verify_serp_parsing(
        "https://web.archive.org/web/20180903061406id_/https://yandex.ru/images/search?p=4&text=speed%20force&img_url=https%3A%2F%2Fwww.old-games.ru%2Fforum%2Fattachments%2Faveg-by_z_2016_01_04_20_06_43_66ae17e6-jpg.98982%2F&pos=132&uinfo=sw-1280-sh-800-ww-1280-wh-649-pd-1-wp-16x10_1280x800&rpt=simage",
        "yandex",
    )


def test_parse_query_yandex_speed_force_1535895408():
    verify_serp_parsing(
        "https://web.archive.org/web/20180902133648id_/https://yandex.ru/images/search?p=15&text=speed%20force&img_url=https%3A%2F%2Fget.wallhere.com%2Fphoto%2Fillustration-anime-death-God-Flash-comics-ART-color-comic-book-fiction-5714.jpg&pos=454&rpt=simage",
        "yandex",
    )



from click import group

from archive_query_log.legacy.cli.alexa import alexa
from archive_query_log.legacy.cli.corpus import corpus_command
from archive_query_log.legacy.cli.external import external
from archive_query_log.legacy.cli.index import index_command
from archive_query_log.legacy.cli.make import make_group


@group()
def main():
    pass


main.add_command(alexa)
main.add_command(corpus_command)
main.add_command(external)
main.add_command(index_command)
main.add_command(make_group)



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_bing_uscis_forms_400_1486690408():
    verify_serp_parsing(
        "https://web.archive.org/web/20170210013328id_/http://www.bing.com/images/search?q=USCIS+Forms+400&FORM=RESTAB",
        "bing",
    )


def test_parse_query_bing_regional_asia_taiwan_localities_taichung_city_1580952978():
    verify_serp_parsing(
        "https://web.archive.org/web/20200206013618id_/https://www.bing.com/search?q=Regional+Asia+Taiwan+Localities+Taichung+City",
        "bing",
    )


def test_parse_query_bing_intel_i7_chip_1656625238():
    verify_serp_parsing(
        "https://web.archive.org/web/20220630214038id_/https://www.bing.com/search?q=intel+i7+chip&lvl=1&FORM=PMETIS&filters=ans%3A%22cvns%22+level%3A%220%22+mcid%3A%229363152adf5f441aa0dd1d28d3cf21f2%22&idx=0",
        "bing",
    )


def test_parse_query_bing_rpg_title_screen_1656104004():
    verify_serp_parsing(
        "https://web.archive.org/web/20220624205324id_/https://www.bing.com/images/search?q=RPG+Title+Screen&FORM=IRMHRS",
        "bing",
    )


def test_parse_query_bing_win10_iso_xia_zai_1580349240():
    verify_serp_parsing(
        "https://web.archive.org/web/20200130015400id_/https://www.bing.com/search?q=win10+iso+%E4%B8%8B%E8%BD%BD&filters=ex1%3A%22ez1%22&qpvt=win10+iso+%E4%B8%8B%E8%BD%BD",
        "bing",
    )


def test_parse_query_bing_great_blue_heron_sound_1660552923():
    verify_serp_parsing(
        "https://web.archive.org/web/20220815084203id_/https://www.bing.com/search?q=great+blue+heron+sound&FORM=QSRE1",
        "bing",
    )


def test_parse_query_bing_wonder_kids_beginning_1652132287():
    verify_serp_parsing(
        "https://web.archive.org/web/20220509213807id_/https://www.bing.com/search?q=wonder+kids+beginning&pc=0TRO&ptag=C1N6612A70619013F5&form=CONBNT&conlogo=CT3210127",
        "bing",
    )


def test_parse_query_bing_mesozoic_wikipedia_1643229373():
    verify_serp_parsing(
        "https://web.archive.org/web/20220126203613id_/https://www.bing.com/search?q=Mesozoic+wikipedia&form=WIKIRE",
        "bing",
    )


def test_parse_query_bing_kumolife_1388896361():
    verify_serp_parsing(
        "https://web.archive.org/web/20140105043241id_/http://www.bing.com/search?q=kumolife&form=MSSRPD",
        "bing",
    )


def test_parse_query_bing_florence_caillon_bandcamp_1647897680():
    verify_serp_parsing(
        "https://web.archive.org/web/20220321212120id_/https://www.bing.com/search?q=Florence+Caillon+bandcamp&qs=n&form=QBRE&=Search+%7B0%7D+for+%7B1%7D&=Search+work+for+%7B0%7D&=%25eManage+Your+Search+History%25E&sp=-1&pq=florence+caillon+bandcamp&sc=1-25&sk=&cvid=E5A3CCFC449440CABF3E217E535C5DD6",
        "bing",
    )


def test_parse_query_bing_common_use_facilities_1643636012():
    verify_serp_parsing(
        "https://web.archive.org/web/20220131133332id_/https://www.bing.com/search?q=Common+use+facilities",
        "bing",
    )


def test_parse_query_bing_corner_app_1618998889():
    verify_serp_parsing(
        "https://web.archive.org/web/20210421095449id_/https://www.bing.com/search?q=Corner+App",
        "bing",
    )


def test_parse_query_bing_blockbench_github_1652298697():
    verify_serp_parsing(
        "https://web.archive.org/web/20220511195137id_/https://www.bing.com/search?q=blockbench+github&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_dash_investment_foundation_1644415217():
    verify_serp_parsing(
        "https://web.archive.org/web/20220209140017id_/https://www.bing.com/search?q=Dash+Investment+Foundation",
        "bing",
    )


def test_parse_query_bing_mehmet_oz_florida_hunting_license_1657728564():
    verify_serp_parsing(
        "https://web.archive.org/web/20220713160924id_/http://www.bing.com/search?q=Mehmet+Oz+Florida+hunting+license&FORM=HDRSC1",
        "bing",
    )


def test_parse_query_bing_agent_077_mission_bloody_mary_movie_1593376539():
    verify_serp_parsing(
        "https://web.archive.org/web/20200628203539id_/https://www.bing.com/search?q=agent+077+mission+bloody+mary+movie&FORM=HDRSC1",
        "bing",
    )


def test_parse_query_bing_connecticut_covid_deaths_1628019384():
    verify_serp_parsing(
        "https://web.archive.org/web/20210803193624id_/https://www.bing.com/search?q=Connecticut+covid+deaths",
        "bing",
    )


def test_parse_query_bing_sungard_1555022695():
    verify_serp_parsing(
        "https://web.archive.org/web/20190411224455id_/https://www.bing.com/search?q=SunGard&filters=ufn%3a%22SunGard%22+sid%3a%22c066bda2-94db-d8c0-eb7a-2ddf1fb005e6%22&FORM=SNAPST",
        "bing",
    )


def test_parse_query_bing_limbpower_1640025328():
    verify_serp_parsing(
        "https://web.archive.org/web/20211220183528id_/https://www.bing.com/search?q=LimbPower",
        "bing",
    )


def test_parse_query_bing_clobetasol_without_prescription_1615178898():
    verify_serp_parsing(
        "https://web.archive.org/web/20210308044818id_/http://www.bing.com/search?q=clobetasol+without+prescription",
        "bing",
    )


def test_parse_query_bing_classement_de_la_premier_league_1658725630():
    verify_serp_parsing(
        "https://web.archive.org/web/20220725050710id_/https://www.bing.com/search?q=classement+de+la+premier+league&FORM=RS1208",
        "bing",
    )


def test_parse_query_bing_crimson_coward_1640789839():
    verify_serp_parsing(
        "https://web.archive.org/web/20211229145719id_/https://www.bing.com/search?q=Crimson+Coward",
        "bing",
    )


def test_parse_query_bing_harukazewakaba0_1643181401():
    verify_serp_parsing(
        "https://web.archive.org/web/20220126071641id_/https://www.bing.com/search?q=harukazewakaba0&FORM=HDRSC1",
        "bing",
    )


def test_parse_query_bing_nexus_mods_manager_1662582175():
    verify_serp_parsing(
        "https://web.archive.org/web/20220907202255id_/https://www.bing.com/search?q=nexus+mods+manager&FORM=QSRE4",
        "bing",
    )


def test_parse_query_bing_shang_chi_film_1563662560():
    verify_serp_parsing(
        "https://web.archive.org/web/20190720224240id_/https://www.bing.com/search?q=Shang-Chi+film&filters=ufn%3a%22Shang-Chi+film%22+sid%3a%2233883806-0a4a-6aae-5598-51112dd2a9f4%22+catguid%3a%22ca394de6-2ef0-65dd-291d-1b2c098e50c3_cfb02057%22+segment%3a%22generic.carousel%22&FORM=SNAPST",
        "bing",
    )


def test_parse_query_bing_obama_speech_1274497739():
    verify_serp_parsing(
        "https://web.archive.org/web/20100522030859id_/http://www.bing.com:80/search?q=Obama+Speech&FORM=QSRE4",
        "bing",
    )


def test_parse_query_bing_elon_musk_tesla_stock_drop_1656685094():
    verify_serp_parsing(
        "https://web.archive.org/web/20220701141814id_/https://www.bing.com/search?q=elon+musk+tesla+stock+drop&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_wikipedia_rock_macabre_1603539954():
    verify_serp_parsing(
        "https://web.archive.org/web/20201024114554id_/https://www.bing.com/search?q=-wikipedia+%22Rock+macabre%22",
        "bing",
    )


def test_parse_query_bing_dan_robitzski_1660531082():
    verify_serp_parsing(
        "https://web.archive.org/web/20220815023802id_/https://www.bing.com/search?q=Dan+Robitzski",
        "bing",
    )


def test_parse_query_bing_arab_suck_and_fuck_bouchra_1365405319():
    verify_serp_parsing(
        "https://web.archive.org/web/20130408071519id_/http://www.bing.com/search?q=arab+suck+and+fuck+bouchra&qs=ns&form=QBLH",
        "bing",
    )


def test_parse_query_bing_wikipedia_orb_univers_de_fiction_1643880033():
    verify_serp_parsing(
        "https://web.archive.org/web/20220203092033id_/https://www.bing.com/search?q=-wikipedia+%22Orb+(univers+de+fiction)%22",
        "bing",
    )


def test_parse_query_bing_vintage_pittsburgh_steelers_steel_curtain_1497160668():
    verify_serp_parsing(
        "https://web.archive.org/web/20170611055748id_/https://www.bing.com/images/search?q=Vintage+Pittsburgh+Steelers+Steel+Curtain&FORM=RESTAB",
        "bing",
    )


def test_parse_query_bing_team_liquid_teams_1652030641():
    verify_serp_parsing(
        "https://web.archive.org/web/20220508172401id_/https://www.bing.com/search?FORM=SNAPST&q=Team+Liquid+teams&filters=ufn%3A%22team+liquid+apex+legends%22+sid%3A%22fb8246be-afdd-9186-3fb4-757eae77b365%22+catguid%3A%224cfd7f5f-fec1-e5d7-15c6-9c6539ed366d_6848f442%22+segment%3A%22generic.carousel%22+gsexp%3A%224cfd7f5f-fec1-e5d7-15c6-9c6539ed366d_bXNvL3Nwb3J0cy5sZWFndWUudGVhbXN8VHJ1ZXxtc28vc3BvcnRzLmxlYWd1ZV9wYXJ0aWNpcGF0aW9uLnRlYW18VHJ1ZQ..%22",
        "bing",
    )


def test_parse_query_bing_european_cannibalism_1591185526():
    verify_serp_parsing(
        "https://web.archive.org/web/20200603115846id_/https://www.bing.com/search?q=european+cannibalism&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_customer_service_phone_number_for_hp_printer_1662986585():
    verify_serp_parsing(
        "https://web.archive.org/web/20220912124305id_/https://www.bing.com/search?q=customer+service+phone+number+for+hp+printer&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_mighty_music_1578065758():
    verify_serp_parsing(
        "https://web.archive.org/web/20200103153558id_/https://www.bing.com/images/search?q=Mighty+Music&FORM=HDRSC2",
        "bing",
    )


def test_parse_query_bing_christian_r_ntizimira_1607730426():
    verify_serp_parsing(
        "https://web.archive.org/web/20201211234706id_/https://www.bing.com/search?q=Christian+R.+Ntizimira",
        "bing",
    )


def test_parse_query_bing_libyan_syrian_union_1643642965():
    verify_serp_parsing(
        "https://web.archive.org/web/20220131152925id_/https://www.bing.com/search?q=Libyan+Syrian+Union",
        "bing",
    )


def test_parse_query_bing_1968_cougar_on_craigslist_1480218715():
    verify_serp_parsing(
        "https://web.archive.org/web/20161127035155id_/http://www.bing.com/search?q=1968+cougar+on+craigslist&form=QBRE",
        "bing",
    )


def test_parse_query_bing_explore_with_us_youtube_emma_1651605205():
    verify_serp_parsing(
        "https://web.archive.org/web/20220503191325id_/https://www.bing.com/search?q=explore+with+us+youtube+emma&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_clay_andrews_1649253287():
    verify_serp_parsing(
        "https://web.archive.org/web/20220406135447id_/https://www.bing.com/search?q=Clay+Andrews",
        "bing",
    )


def test_parse_query_bing_kutuklui_affair_1643643819():
    verify_serp_parsing(
        "https://web.archive.org/web/20220131154339id_/https://www.bing.com/search?q=Kutuklui+Affair",
        "bing",
    )


def test_parse_query_bing_the_clavarioid_fungi_of_new_zealand_ron_petersen_1556393373():
    verify_serp_parsing(
        "https://web.archive.org/web/20190427192933id_/https://www.bing.com/search?q=The+clavarioid+fungi+of+New+Zealand+Ron+Petersen&filters=ufn%3a%22The+clavarioid+fungi+of+New+Zealand+Ron+Petersen%22+sid%3a%220b30c0ac-f55d-3966-3df7-6c37ffc7e507%22+catguid%3a%229d080d55-a2a0-4c60-debe-b536d8a0aa82_c0bcbbc4%22+segment%3a%22generic.carousel%22&FORM=SNAPST",
        "bing",
    )


def test_parse_query_bing_wikipedia_port_de_gennevilliers_metro_de_paris_1604443017():
    verify_serp_parsing(
        "https://web.archive.org/web/20201103223657id_/https://www.bing.com/search?q=-wikipedia+%22Port+de+Gennevilliers+(m%C3%A9tro+de+Paris)%22",
        "bing",
    )


def test_parse_query_bing_wikipedia_vesuvius_1641311740():
    verify_serp_parsing(
        "https://web.archive.org/web/20220104155540id_/https://www.bing.com/search?q=-wikipedia+%22Vesuvius%22",
        "bing",
    )


def test_parse_query_bing_countercyclical_fiscal_policy_1643622693():
    verify_serp_parsing(
        "https://web.archive.org/web/20220131095133id_/https://www.bing.com/search?q=Countercyclical+fiscal+policy",
        "bing",
    )


def test_parse_query_bing_herzlich_willkommen_unterbindet_den_storeinfluss_bei_psychosen_1650332555():
    verify_serp_parsing(
        "https://web.archive.org/web/20220419014235id_/https://www.bing.com/search?q=%22Herzlich+willkommen+!%22+unterbindet+den+St%C3%B6reinfluss+bei+Psychosen&go=Suche&qs=ds&form=QBRE",
        "bing",
    )


def test_parse_query_bing_limited_range_icbm_1643649846():
    verify_serp_parsing(
        "https://web.archive.org/web/20220131172406id_/https://www.bing.com/search?q=Limited+Range+ICBM",
        "bing",
    )


def test_parse_query_bing_lake_wappapello_state_wildlife_area_1643615402():
    verify_serp_parsing(
        "https://web.archive.org/web/20220131075002id_/https://www.bing.com/search?q=Lake+Wappapello+State+Wildlife+Area",
        "bing",
    )


def test_parse_query_bing_sweden_riots_1618569435():
    verify_serp_parsing(
        "https://web.archive.org/web/20210416103715id_/https://www.bing.com/search?q=Sweden+Riots&FORM=HDRSC1",
        "bing",
    )


def test_parse_query_bing_dataquantum_com_vancouver_hosting_1375879464():
    verify_serp_parsing(
        "https://web.archive.org/web/20130807124424id_/http://www.bing.com/search?q=%64%61%74%61%71%75%61%6e%74%75%6d%2e%63%6f%6d%2b%56%61%6e%63%6f%75%76%65%72%2b%68%6f%73%74%69%6e%67",
        "bing",
    )


def test_parse_query_bing_cody_machan_facebook_1594008363():
    verify_serp_parsing(
        "https://web.archive.org/web/20200706040603id_/https://www.bing.com/search?q=cody+machan+facebook&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_allovercr_1423275634():
    verify_serp_parsing(
        "https://web.archive.org/web/20150207022034id_/http://www.bing.com/search?q=allovercr&go=Enviar+consulta&qs=ds&form=QBRE",
        "bing",
    )


def test_parse_query_bing_aldershot_news_1629564105():
    verify_serp_parsing(
        "https://web.archive.org/web/20210821164145id_/https://www.bing.com/search?q=aldershot+news&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_eero_pro_mesh_wifi_system_1659437763():
    verify_serp_parsing(
        "https://web.archive.org/web/20220802105603id_/https://www.bing.com/search?q=eero+pro+mesh+wifi+system&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_computers_software_presentation_1488623366():
    verify_serp_parsing(
        "https://web.archive.org/web/20170304102926id_/https://www.bing.com/search?q=Computers+Software+Presentation",
        "bing",
    )


def test_parse_query_bing_1137ce_1639788231():
    verify_serp_parsing(
        "https://web.archive.org/web/20211218004351id_/http://www.bing.com/search?q=1137CE",
        "bing",
    )


def test_parse_query_bing_wikipedia_sider_tribu_1593501017():
    verify_serp_parsing(
        "https://web.archive.org/web/20200630071017id_/https://www.bing.com/search?q=-wikipedia+%22Sider+(tribu)%22",
        "bing",
    )


def test_parse_query_bing_100_9_classic_country_radio_station_alaska_1660338459():
    verify_serp_parsing(
        "https://web.archive.org/web/20220812210739id_/https://www.bing.com/search?q=100.9+classic+country+radio+station+alaska&FORM=QSRE6",
        "bing",
    )


def test_parse_query_bing_wikipedia_privatbrauerei_hoepfner_1620534005():
    verify_serp_parsing(
        "https://web.archive.org/web/20210509042005id_/https://www.bing.com/search?q=-wikipedia+%22Privatbrauerei+Hoepfner%22",
        "bing",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_tribunnews_peringatan_dini_bmkg_kamis_18_februari_2021_1632679908():
    verify_serp_parsing(
        "https://web.archive.org/web/20210926181148id_/https://www.tribunnews.com/search?q=peringatan+dini+bmkg+kamis+18+februari+2021",
        "tribunnews",
    )


def test_parse_query_tribunnews_search_term_string_1607106977():
    verify_serp_parsing(
        "https://web.archive.org/web/20201204183617id_/https://www.tribunnews.com/search?q={search_term_string}",
        "tribunnews",
    )


def test_parse_query_tribunnews_search_term_string_1607225592():
    verify_serp_parsing(
        "https://web.archive.org/web/20201206033312id_/https://www.tribunnews.com/search?q={search_term_string}",
        "tribunnews",
    )


def test_parse_query_tribunnews_search_term_string_1607224881():
    verify_serp_parsing(
        "https://web.archive.org/web/20201206032121id_/https://www.tribunnews.com/search?q={search_term_string}",
        "tribunnews",
    )


def test_parse_query_tribunnews_search_term_string_1607130238():
    verify_serp_parsing(
        "https://web.archive.org/web/20201205010358id_/https://www.tribunnews.com/search?q={search_term_string}",
        "tribunnews",
    )


def test_parse_query_tribunnews_polsek_tallo_1663047668():
    verify_serp_parsing(
        "https://web.archive.org/web/20220913054108id_/http://www.tribunnews.com/search?q=polsek+tallo",
        "tribunnews",
    )


def test_parse_query_tribunnews_ricky_natapradja_1663049121():
    verify_serp_parsing(
        "https://web.archive.org/web/20220913060521id_/http://www.tribunnews.com/search?q=ricky+natapradja",
        "tribunnews",
    )


def test_parse_query_tribunnews_pilgub_dki_jakarta_2017_1491952670():
    verify_serp_parsing(
        "https://web.archive.org/web/20170411231750id_/http://www.tribunnews.com:80/search?q=pilgub+dki+jakarta+2017",
        "tribunnews",
    )


def test_parse_query_tribunnews_search_term_string_1607120098():
    verify_serp_parsing(
        "https://web.archive.org/web/20201204221458id_/https://www.tribunnews.com/search?q={search_term_string}",
        "tribunnews",
    )


def test_parse_query_tribunnews_linkid_js_1533168227():
    verify_serp_parsing(
        "https://web.archive.org/web/20180802000347id_/http://www.tribunnews.com/search?q=linkid.js",
        "tribunnews",
    )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_naver_11548566_973003263():
    verify_serp_parsing(
        "https://web.archive.org/web/20001031144103id_/http://search.naver.com:80/search.naver?where=web&type=3&query=11548566",
        "naver",
    )


def test_parse_query_naver_monteria_montheria_bunyupoteu_jeongitipoteu_jeongijujeonja_bunyupoteugi_1200ml_1200mlpingkeu_hugi_1640713881():
    verify_serp_parsing(
        "https://web.archive.org/web/20211228175121id_/https://search.naver.com/search.naver?query=%EB%AA%AC%ED%85%8C%EB%A6%AC%EC%95%84%20Montheria%20%EB%B6%84%EC%9C%A0%ED%8F%AC%ED%8A%B8%20%EC%A0%84%EA%B8%B0%ED%8B%B0%ED%8F%AC%ED%8A%B8%20%EC%A0%84%EA%B8%B0%EC%A3%BC%EC%A0%84%EC%9E%90%20%EB%B6%84%EC%9C%A0%ED%8F%AC%ED%8A%B8%EA%B8%B0%201200ML,%201200ml%ED%95%91%ED%81%AC+%ED%9B%84%EA%B8%B0",
        "naver",
    )


def test_parse_query_naver_heolkeu_dari_gajin_namja_1397449417():
    verify_serp_parsing(
        "https://web.archive.org/web/20140414042337id_/http://search.naver.com/search.naver?where=nexearch&query=%ED%97%90%ED%81%AC+%EB%8B%A4%EB%A6%AC+%EA%B0%80%EC%A7%84+%EB%82%A8%EC%9E%90&ie=utf8&sm=nws_htk",
        "naver",
    )


def test_parse_query_naver_10559239_974530320():
    verify_serp_parsing(
        "https://web.archive.org/web/20001118065200id_/http://search.naver.com:80/search.naver?where=web&type=3&query=10559239",
        "naver",
    )


def test_parse_query_naver_junggangosagongbu_1627424756():
    verify_serp_parsing(
        "https://web.archive.org/web/20210727222556id_/https://search.naver.com/search.naver?where=view&sm=tab_viw.all&query=%EC%A4%91%EA%B0%84%EA%B3%A0%EC%82%AC%EA%B3%B5%EB%B6%80",
        "naver",
    )


def test_parse_query_naver_dpsxmfl_1632292908():
    verify_serp_parsing(
        "https://web.archive.org/web/20210922064148id_/https://search.naver.com/search.naver?where=nexearch&query=dpsxmfl&x=31&y=5&frm=t1",
        "naver",
    )


def test_parse_query_naver_sejongmunhwahoegwan_daegeugjang_1652993423():
    verify_serp_parsing(
        "https://web.archive.org/web/20220519205023id_/https://search.naver.com/search.naver?where=image&sm=tab_jum&query=%EC%84%B8%EC%A2%85%EB%AC%B8%ED%99%94%ED%9A%8C%EA%B4%80+%EB%8C%80%EA%B7%B9%EC%9E%A5",
        "naver",
    )


def test_parse_query_naver_geobugseon_1439284748():
    verify_serp_parsing(
        "https://web.archive.org/web/20150811091908id_/http://search.naver.com/search.naver?where=nexearch&query=%EA%B1%B0%EB%B6%81%EC%84%A0&sm=top_lve&ie=utf8",
        "naver",
    )


def test_parse_query_naver_z_974687640():
    verify_serp_parsing(
        "https://web.archive.org/web/20001120023400id_/http://search.naver.com:80/search.naver?where=nexearch&query=%C0%E5%BF%DC+%B9%DF%B8%C5%BC%D2",
        "naver",
    )


def test_parse_query_naver_t_1047137339():
    verify_serp_parsing(
        "https://web.archive.org/web/20030308152859id_/http://search.naver.com:80/search.naver?where=nexearch&query=%bd%ba%c6%ae%b8%b3%bc%ee%ba%b8%bf%a9%c1%e0",
        "naver",
    )



from argparse import ArgumentParser
from datetime import datetime
from gzip import GzipFile
from json import loads, JSONDecodeError, dumps
from pathlib import Path
from random import shuffle
from typing import Optional, Iterator, Literal
from urllib.parse import urlparse
from uuid import uuid5, NAMESPACE_URL

from fastwarc import FileStream, ArchiveIterator, WarcRecordType, WarcRecord
from publicsuffixlist import PublicSuffixList
from pyspark.sql import SparkSession
from tqdm.auto import tqdm
from yaml import safe_load

_CEPH_DIR = Path("/mnt/ceph/storage")
_RESEARCH_DIR = _CEPH_DIR / "data-in-progress" / "data-research"
_GLOBAL_DATA_DIR = _RESEARCH_DIR / "web-search" / "web-archive-query-log"
_DATA_DIR = _GLOBAL_DATA_DIR / "focused"

_PUBLIC_SUFFIX_LIST = PublicSuffixList()


def _load_services(path: Path) -> dict:
    with path.open("r") as file:
        services_dict = safe_load(file)
    services_list = [(service["name"], service) for service in services_dict]
    assert len({name for name, service in services_list}) == len(services_list)
    services = {
        name: service
        for name, service in services_list
    }
    print(f"Found {len(services)} services.")
    return services


_SERVICES = _load_services(_GLOBAL_DATA_DIR / "selected-services.yaml")


def _detect_language(text: str) -> Optional[str]:
    text = text.replace("\n", " ")
    from cld3 import get_language
    language_prediction = get_language(text)
    if language_prediction is None:
        return None
    return language_prediction.language.split("-")[0] \
        if language_prediction.is_reliable else None


def _read_jsonl(path: Path, base_type: str) -> dict:
    jsonl_path = _DATA_DIR / base_type / path.with_suffix(".jsonl.gz")
    if not jsonl_path.exists():
        return {}
    index = {}
    try:
        with GzipFile(jsonl_path, "r") as gzip_file:
            # noinspection PyTypeChecker
            for line in tqdm(gzip_file, desc="Index JSONL"):
                try:
                    # noinspection PyTypeChecker
                    record = loads(line)
                except:
                    print(f"Could not index {line} at {path}.")
                    continue
                record_id = uuid5(
                    NAMESPACE_URL,
                    f"{record['timestamp']}:{record['url']}",
                )
                index[record_id] = record
        return index
    except:
        print(f"Could not read JSONL file at {path}.")
        return {}


def _index_warc(path: Path, base_type: str) -> dict:
    warc_path = _DATA_DIR / base_type / path
    if not warc_path.exists():
        return {}
    index = {}
    for warc_child_path in warc_path.iterdir():
        if warc_child_path.name.startswith("."):
            continue
        try:
            stream = FileStream(str(warc_child_path.absolute()))
            records = ArchiveIterator(
                stream,
                record_types=WarcRecordType.response,
                parse_http=False,
            )
            # noinspection PyTypeChecker
            for record in tqdm(records, desc="Index WARC"):
                record: WarcRecord
                offset = record.stream_pos
                record_url_header = record.headers["Archived-URL"]
                try:
                    record_url = loads(record_url_header)
                except JSONDecodeError:
                    print(f"Could not index {record_url_header} at {path}.")
                    continue
                record_id = uuid5(
                    NAMESPACE_URL,
                    f"{record_url['timestamp']}:{record_url['url']}",
                )
                index[record_id] = (
                    warc_child_path,
                    offset,
                )
        except:
            print(f"Could not read WARC file at {warc_child_path}.")
            continue
    return index


def _iter_relative_path_records(relative_path: Path) -> Iterator[tuple]:
    service = relative_path.parts[0]

    print(f"Reading files in {relative_path}.")
    archived_urls_index = _read_jsonl(relative_path, "archived-urls")
    print("Finished reading archived URLs.")
    archived_query_urls_index = _read_jsonl(relative_path,
                                            "archived-query-urls")
    print("Finished reading archived query URLs.")
    archived_raw_serps_index = _index_warc(relative_path, "archived-raw-serps")
    print("Finished reading archived raw SERPs (pointers).")
    archived_parsed_serps_index = _read_jsonl(relative_path,
                                              "archived-parsed-serps")
    print("Finished reading archived parsed SERPs.")

    for record_id, archived_url in archived_urls_index.items():
        archived_url = archived_urls_index[record_id]
        archived_query_url = archived_query_urls_index.get(record_id, None)
        archived_raw_serp_location = archived_raw_serps_index.get(record_id,
                                                                  None)
        archived_parsed_serp = archived_parsed_serps_index.get(record_id, None)

        yield service, relative_path, record_id, archived_url, archived_query_url, archived_raw_serp_location, archived_parsed_serp


def _iter_results(
        archived_url: dict, archived_parsed_serp: dict
) -> Iterator[dict]:
    if archived_parsed_serp is None:
        return

    for snippet in archived_parsed_serp["results"]:
        url = snippet["url"]
        domain = urlparse(url).hostname
        public_suffix = _PUBLIC_SUFFIX_LIST.publicsuffix(domain) \
            if domain is not None else None
        timestamp = archived_url["timestamp"]
        wayback_timestamp = \
            datetime.fromtimestamp(timestamp).strftime("%Y%m%d%H%M%S")
        wayback_url = f"https://web.archive.org/web/{wayback_timestamp}/{url}"
        wayback_raw_url = \
            f"https://web.archive.org/web/{wayback_timestamp}id_/{url}"
        record_id = uuid5(
            NAMESPACE_URL,
            f"{snippet['rank']}:{snippet['timestamp']}:{snippet['url']}"
        )
        print(f"Yield result: {record_id}")
        yield {
            "result_id": str(record_id),
            "result_url": url,
            "result_domain": domain,
            "result_domain_public_suffix": public_suffix,
            "result_wayback_url": wayback_url,
            "result_wayback_raw_url": wayback_raw_url,
            "result_snippet_rank": snippet['rank'],
            "result_snippet_title": snippet["title"],
            "result_snippet_text": snippet["snippet"],
            "result_warc_relative_path": None,
            "result_warc_byte_offset": None,
        }


def _record_to_query(relative_path_record: tuple) -> Optional[str]:
    service, relative_path, record_id, archived_url, archived_query_url, \
        archived_raw_serp_location, archived_parsed_serp = relative_path_record

    if archived_query_url is None:
        print(f"Archived query URL not found for ID {record_id}.")
        return None

    url = archived_url["url"]
    domain = urlparse(url).hostname
    public_suffix = _PUBLIC_SUFFIX_LIST.publicsuffix(domain) \
        if domain is not None else None
    timestamp = archived_url["timestamp"]
    wayback_timestamp = \
        datetime.fromtimestamp(timestamp).strftime("%Y%m%d%H%M%S")
    wayback_url = f"https://web.archive.org/web/{wayback_timestamp}/{url}"
    wayback_raw_url = \
        f"https://web.archive.org/web/{wayback_timestamp}id_/{url}"
    query = archived_query_url["query"]
    language = _detect_language(query)
    service_info = _SERVICES[service]

    documents = list(_iter_results(archived_url, archived_parsed_serp))

    print(f"Yield SERP: {record_id}")
    return dumps({
        "serp_id": str(record_id),
        "serp_url": url,
        "serp_domain": domain,
        "serp_domain_public_suffix": public_suffix,
        "serp_timestamp": timestamp,
        "serp_wayback_url": wayback_url,
        "serp_wayback_raw_url": wayback_raw_url,
        "serp_page": archived_query_url["page"],
        "serp_offset": archived_query_url["offset"],
        "serp_query_text_url": query,
        "serp_query_text_url_language": language,
        "serp_query_text_html": (
            archived_parsed_serp["interpreted_query"]
            if archived_parsed_serp is not None else None
        ),
        "serp_warc_relative_path": (
            str(archived_raw_serp_location[0].relative_to(_GLOBAL_DATA_DIR))
            if archived_raw_serp_location is not None else None
        ),
        "serp_warc_byte_offset": (
            archived_raw_serp_location[1]
            if archived_raw_serp_location is not None else None
        ),
        "serp_results": documents,
        "search_provider_name": service,
        "search_provider_alexa_domain": service_info["alexa_domain"],
        "search_provider_alexa_domain_public_suffix":
            _SERVICES[service]["public_suffix"],
        "search_provider_alexa_rank": service_info["alexa_rank"],
        "search_provider_category": service_info["category"],
    })


def _iter_query_documents(query: str) -> Iterator[dict]:
    query = loads(query)
    for result in query["serp_results"]:
        print(f"Yield result: {result['result_id']}")
        yield dumps({
            "result_id": result["result_id"],
            "result_url": result["result_url"],
            "result_domain": result["result_domain"],
            "result_domain_public_suffix":
                result["result_domain_public_suffix"],
            "result_wayback_url": result["result_wayback_url"],
            "result_wayback_raw_url": result["result_wayback_raw_url"],
            "result_snippet_rank": result["result_snippet_rank"],
            "result_snippet_title": result["result_snippet_title"],
            "result_snippet_text": result["result_snippet_text"],
            "result_warc_relative_path": result["result_warc_relative_path"],
            "result_warc_byte_offset": result["result_warc_byte_offset"],
            "serp_id": query["serp_id"],
            "serp_url": query["serp_url"],
            "serp_domain": query["serp_domain"],
            "serp_domain_public_suffix": query["serp_domain_public_suffix"],
            "serp_timestamp": query["serp_timestamp"],
            "serp_wayback_url": query["serp_wayback_url"],
            "serp_wayback_raw_url": query["serp_wayback_raw_url"],
            "serp_page": query["serp_page"],
            "serp_offset": query["serp_offset"],
            "serp_query_text_url": query["serp_query_text_url"],
            "serp_query_text_url_language":
                query["serp_query_text_url_language"],
            "serp_query_text_html": query["serp_query_text_html"],
            "serp_warc_relative_path": query["serp_warc_relative_path"],
            "serp_warc_byte_offset": query["serp_warc_byte_offset"],
            "search_provider_name": query["search_provider_name"],
            "search_provider_alexa_domain":
                query["search_provider_alexa_domain"],
            "search_provider_alexa_domain_public_suffix":
                query["search_provider_alexa_domain_public_suffix"],
            "search_provider_alexa_rank": query["search_provider_alexa_rank"],
            "search_provider_category": query["search_provider_category"],
        })


def main(variant: Literal["small", "medium", "full"]):
    session = SparkSession.builder.getOrCreate()

    sc = session.sparkContext

    relative_paths = [
        path
        .relative_to(_DATA_DIR / "archived-urls")
        .with_name(path.name[:-len(".jsonl.gz")])
        for path in _DATA_DIR.glob("archived-urls/*/*/*.jsonl.gz")
    ]
    print(f"Found {len(relative_paths)} paths.")
    shuffle(relative_paths)
    if variant == "small":
        relative_paths = relative_paths[:100]
    elif variant == "medium":
        relative_paths = relative_paths[:2500]
    print(f"Selected {len(relative_paths)} paths for corpus creation.")

    print("Start corpus creation.")

    print("Export SERPs.")
    sc.parallelize(relative_paths) \
        .repartition(1_000) \
        .flatMap(_iter_relative_path_records) \
        .map(_record_to_query) \
        .filter(lambda json: json is not None) \
        .saveAsTextFile(f"archive-query-log/{variant}/serps/",
                        compressionCodecClass=
                        "org.apache.hadoop.io.compress.GzipCodec")

    print("Export results.")
    sc.textFile(f"archive-query-log/{variant}/serps/") \
        .flatMap(_iter_query_documents) \
        .map(dumps) \
        .repartition(1_000) \
        .saveAsTextFile(f"archive-query-log/{variant}/results/",
                        compressionCodecClass=
                        "org.apache.hadoop.io.compress.GzipCodec")

    print("Done.")


if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument(
        "--small",
        dest="variant",
        action="store_const",
        const="small",
        default="full",
    )
    parser.add_argument(
        "--medium",
        dest="variant",
        action="store_const",
        const="medium",
        default="full",
    )
    parser.add_argument(
        "--full",
        dest="variant",
        action="store_const",
        const="full",
        default="full",
    )
    args = parser.parse_args()
    print(f"Creating corpus (variant: {args.variant}).")
    main(args.variant)



from functools import cache
from itertools import chain, islice
from typing import Iterable, Iterator
from uuid import uuid5

from click import echo
from elasticsearch_dsl import Search
from elasticsearch_dsl.function import RandomScore
from elasticsearch_dsl.query import FunctionScore, Term, RankFeature, Exists
from tqdm.auto import tqdm

from archive_query_log.config import Config
from archive_query_log.namespaces import NAMESPACE_URL_QUERY_PARSER
from archive_query_log.orm import (
    Capture,
    Serp,
    InnerCapture,
    InnerParser,
    UrlQueryParser,
)
from archive_query_log.orm import UrlQueryParserType, InnerProviderId
from archive_query_log.parsers.url import (
    parse_url_query_parameter,
    parse_url_fragment_parameter,
    parse_url_path_segment,
)
from archive_query_log.parsers.util import clean_text
from archive_query_log.utils.es import safe_iter_scan, update_action
from archive_query_log.utils.time import utc_now


def add_url_query_parser(
    config: Config,
    provider_id: str | None,
    url_pattern_regex: str | None,
    priority: float | None,
    parser_type: UrlQueryParserType,
    parameter: str | None,
    segment: int | None,
    remove_pattern_regex: str | None,
    space_pattern_regex: str | None,
) -> None:
    if priority is not None and priority <= 0:
        raise ValueError("Priority must be strictly positive.")
    if parser_type == "query_parameter":
        if parameter is None:
            raise ValueError("No query parameter given.")
    elif parser_type == "fragment_parameter":
        if parameter is None:
            raise ValueError("No fragment parameter given.")
    elif parser_type == "path_segment":
        if segment is None:
            raise ValueError("No path segment given.")
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    parser_id_components = (
        provider_id if provider_id is not None else "",
        url_pattern_regex if url_pattern_regex is not None else "",
        str(priority) if priority is not None else "",
    )
    parser_id = str(
        uuid5(
            NAMESPACE_URL_QUERY_PARSER,
            ":".join(parser_id_components),
        )
    )
    parser = UrlQueryParser(
        id=parser_id,
        last_modified=utc_now(),
        provider=InnerProviderId(id=provider_id) if provider_id else None,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type,
        parameter=parameter,
        segment=segment,
        remove_pattern_regex=remove_pattern_regex,
        space_pattern_regex=space_pattern_regex,
    )
    parser.save(using=config.es.client, index=config.es.index_url_query_parsers)


def _parse_url_query(parser: UrlQueryParser, capture_url: str) -> str | None:
    # Check if URL matches pattern.
    if parser.url_pattern is not None and not parser.url_pattern.match(capture_url):
        return None

    # Parse query.
    if parser.parser_type == "query_parameter":
        if parser.parameter is None:
            raise ValueError("No query parameter given.")
        query = parse_url_query_parameter(parser.parameter, capture_url)
        if query is None:
            return None
        return clean_text(
            text=query,
            remove_pattern=parser.remove_pattern,
            space_pattern=parser.space_pattern,
        )
    elif parser.parser_type == "fragment_parameter":
        if parser.parameter is None:
            raise ValueError("No fragment parameter given.")
        query = parse_url_fragment_parameter(parser.parameter, capture_url)
        if query is None:
            return None
        return clean_text(
            text=query,
            remove_pattern=parser.remove_pattern,
            space_pattern=parser.space_pattern,
        )
    elif parser.parser_type == "path_segment":
        if parser.segment is None:
            raise ValueError("No path segment given.")
        query = parse_url_path_segment(parser.segment, capture_url)
        if query is None:
            return None
        return clean_text(
            text=query,
            remove_pattern=parser.remove_pattern,
            space_pattern=parser.space_pattern,
        )
    else:
        raise ValueError(f"Unknown parser type: {parser.parser_type}")


@cache
def _url_query_parsers(
    config: Config,
    provider_id: str,
) -> list[UrlQueryParser]:
    parsers: Iterable[UrlQueryParser] = (
        UrlQueryParser.search(
            using=config.es.client, index=config.es.index_url_query_parsers
        )
        .filter(~Exists(field="provider.id") | Term(provider__id=provider_id))
        .query(RankFeature(field="priority", saturation={}))
        .scan()
    )
    parsers = safe_iter_scan(parsers)
    return list(parsers)


def _parse_serp_url_query_action(
    config: Config,
    capture: Capture,
) -> Iterator[dict]:
    # Re-check if parsing is necessary.
    if (
        capture.url_query_parser is not None
        and capture.url_query_parser.should_parse is not None
        and not capture.url_query_parser.should_parse
    ):
        return

    for parser in _url_query_parsers(config, capture.provider.id):
        # Try to parse the query.
        url_query = _parse_url_query(parser, capture.url)
        if url_query is None:
            # Parsing was not successful, e.g., URL pattern did not match.
            continue
        serp = Serp(
            id=capture.id,
            last_modified=utc_now(),
            archive=capture.archive,
            provider=capture.provider,
            capture=InnerCapture(
                id=capture.id,
                url=capture.url,
                timestamp=capture.timestamp,
                status_code=capture.status_code,
                digest=capture.digest,
                mimetype=capture.mimetype,
            ),
            url_query=url_query,
            url_query_parser=InnerParser(
                id=parser.id,
                should_parse=False,
                last_parsed=utc_now(),
            ),
            url_page_parser=InnerParser(
                should_parse=True,
            ),
            url_offset_parser=InnerParser(
                should_parse=True,
            ),
            warc_query_parser=InnerParser(
                should_parse=True,
            ),
            warc_snippets_parser=InnerParser(
                should_parse=True,
            ),
        )
        serp.meta.index = config.es.index_serps
        yield serp.to_dict(include_meta=True)
        yield update_action(
            capture,
            url_query_parser=InnerParser(
                id=parser.id,
                should_parse=False,
                last_parsed=utc_now(),
            ),
        )
        return
    yield update_action(
        capture,
        url_query_parser=InnerParser(
            should_parse=False,
            last_parsed=utc_now(),
        ),
    )
    return


def parse_serps_url_query(config: Config, prefetch_limit: int | None = None) -> None:
    config.es.client.indices.refresh(index=config.es.index_captures)
    changed_captures_search: Search = (
        Capture.search(using=config.es.client, index=config.es.index_captures)
        .filter(~Term(url_query_parser__should_parse=False))
        .query(
            RankFeature(field="archive.priority", saturation={})
            | RankFeature(field="provider.priority", saturation={})
            | FunctionScore(functions=[RandomScore()])
        )
    )
    num_changed_captures = changed_captures_search.count()
    if num_changed_captures > 0:
        changed_captures: Iterable[Capture] = changed_captures_search.params(
            preserve_order=True
        ).scan()
        changed_captures = safe_iter_scan(changed_captures)

        if prefetch_limit is not None:
            num_changed_captures = min(num_changed_captures, prefetch_limit)
            changed_captures = tqdm(changed_captures, total=num_changed_captures, desc="Pre-fetching captures", unit="capture")
            changed_captures = iter(list(islice(changed_captures, prefetch_limit)))

        # noinspection PyTypeChecker
        changed_captures = tqdm(
            changed_captures,
            total=num_changed_captures,
            desc="Parsing URL query",
            unit="capture",
        )
        actions = chain.from_iterable(
            _parse_serp_url_query_action(config, capture)
            for capture in changed_captures
        )
        config.es.bulk(actions)
    else:
        echo("No new/changed captures.")



from datetime import datetime
from gzip import open as gzip_open
from typing import NamedTuple, Type
from pathlib import Path

from elasticsearch_dsl.query import Exists, Query, Term
from expiringdict import ExpiringDict
from flask import render_template, Response, make_response
from warcio import ArchiveIterator

from archive_query_log.config import Config
from archive_query_log.orm import (
    Archive,
    Provider,
    Source,
    Capture,
    BaseDocument,
    Serp,
    Result,
    UrlQueryParser,
    UrlPageParser,
    UrlOffsetParser,
    WarcQueryParser,
    WarcSnippetsParser,
)
from archive_query_log.utils.time import utc_now

_CACHE_SECONDS_STATISTICS = 60 * 5  # 5 minutes
_CACHE_SECONDS_WARC_CACHE_STATISTICS = 60 * 1  # 1 minute
_CACHE_SECONDS_PROGRESS = 60 * 10  # 10 minutes


class Statistics(NamedTuple):
    name: str
    description: str
    total: int
    disk_size: str | None
    last_modified: datetime | None


class Progress(NamedTuple):
    input_name: str
    output_name: str
    description: str
    total: int
    current: int


DocumentType = Type[BaseDocument]

_statistics_cache: dict[
    tuple[DocumentType, str, str],
    Statistics,
] = ExpiringDict(
    max_len=100,
    max_age_seconds=_CACHE_SECONDS_STATISTICS,
)


def _convert_bytes(bytes_count: int) -> str:
    step_unit = 1000.0
    bytes_count_decimal: float = bytes_count
    for unit in ["B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB", "RB"]:
        if bytes_count_decimal < step_unit:
            return f"{bytes_count_decimal:3.1f} {unit}"
        bytes_count_decimal /= step_unit
    return f"{bytes_count_decimal:3.1f} QB"


def _get_statistics(
    config: Config,
    name: str,
    description: str,
    index: str,
    document: DocumentType,
    last_modified_field: str = "last_modified"
) -> Statistics:
    key = (document, index, last_modified_field)
    if key in _statistics_cache:
        return _statistics_cache[key]

    config.es.client.indices.refresh(index=index)
    stats = config.es.client.indices.stats(index=index)
    search = document.search(using=config.es.client, index=index)
    search = search.filter(Exists(field=last_modified_field))
    total = search.count()
    last_modified_response = (
        search.sort(f"-{last_modified_field}")
        .extra(size=1)
        .execute()
    )
    if last_modified_response.hits.total.value == 0:
        last_modified = None
    else:
        last_modified = last_modified_response.hits[0].last_modified

    disk_size = (
        _convert_bytes(stats["_all"]["total"]["store"]["size_in_bytes"])
        if last_modified_field == "last_modified"
        else None
    )

    statistics = Statistics(
        name=name,
        description=description,
        total=total,
        disk_size=disk_size,
        last_modified=last_modified,
    )
    _statistics_cache[key] = statistics
    return statistics

_warc_cache_statistics_cache: dict[
    tuple[Path, bool],
    Statistics,
] = ExpiringDict(
    max_len=100,
    max_age_seconds=_CACHE_SECONDS_WARC_CACHE_STATISTICS,
)

def _get_warc_cache_statistics(
    config: Config,
    name: str,
    description: str,
    cache_path: Path,
    temporary: bool = False
) -> Statistics:
    """Retrieve WARC cache statistics."""
    key = (cache_path.resolve(), temporary)
    if key in _warc_cache_statistics_cache:
        return _warc_cache_statistics_cache[key]
    
    file_paths: list[Path]
    if temporary:
        file_paths = list(cache_path.glob(".*.warc.gz"))

    else:
        file_paths = list(cache_path.glob("[!.]*.warc.gz"))

    disk_size_bytes: int = 0
    last_modified: float | None = None
    warc_count: int = 0

    if len(file_paths) > 0:
        disk_size_bytes = sum(file_path.stat().st_size for file_path in file_paths)
        last_modified = max(file_path.stat().st_mtime for file_path in file_paths)
        for file_path in file_paths:
            with gzip_open(file_path, mode="rb") as gzip_file:
                iterator = ArchiveIterator(
                    fileobj=gzip_file,
                    no_record_parse=True,
                )
                warc_count += sum(
                    1 for record in iterator if record.rec_type == "request"
                )

    statistics = Statistics(
        name=name,
        description=description,
        total=warc_count,
        disk_size=_convert_bytes(disk_size_bytes),
        last_modified=datetime.fromtimestamp(last_modified) if last_modified else None,
    )
    _warc_cache_statistics_cache[key] = statistics
    return statistics


_progress_cache: dict[
    tuple[DocumentType, str, str, str],
    Progress,
] = ExpiringDict(
    max_len=100,
    max_age_seconds=_CACHE_SECONDS_PROGRESS,
)


def _get_processed_progress(
    config: Config,
    input_name: str,
    output_name: str,
    description: str,
    document: DocumentType,
    index: str,
    status_field: str,
    filter_query: Query | None = None,
) -> Progress:
    key = (document, index, repr(filter_query), status_field)
    if key in _progress_cache:
        return _progress_cache[key]

    config.es.client.indices.refresh(index=index)
    search = document.search(using=config.es.client, index=index)
    if filter_query is not None:
        search = search.filter(filter_query)
    total = search.count()
    search_processed = search.filter(Term(**{status_field: False}))
    total_processed = search_processed.count()
    progress = Progress(
        input_name=input_name,
        output_name=output_name,
        description=description,
        total=total,
        current=total_processed,
    )
    _progress_cache[key] = progress
    return progress


def home(config: Config) -> str | Response:
    statistics_list: list[Statistics] = [
        _get_statistics(
            config=config,
            name="Archives",
            description="Web archiving services that offer CDX and Memento APIs.",
            document=Archive,
            index=config.es.index_archives,
        ),
        _get_statistics(
            config=config,
            name="Providers",
            description="Search providers, i.e., websites that offer "
            "a search functionality.",
            document=Provider,
            index=config.es.index_providers,
        ),
        _get_statistics(
            config=config,
            name="Sources",
            description="The cross product of all archives and "
            "the provider's domains and URL prefixes.",
            document=Source,
            index=config.es.index_sources,
        ),
        _get_statistics(
            config=config,
            name="Captures",
            description="Captures matching from the archives "
            "that match domain and URL prefixes.",
            document=Capture,
            index=config.es.index_captures,
        ),
        _get_statistics(
            config=config,
            name="SERPs",
            description="Search engine result pages that have been "
            "identified among the captures.",
            document=Serp,
            index=config.es.index_serps,
        ),
        _get_statistics(
            config=config,
            name="+ URL query",
            description="SERPs for which the query has been parsed from the URL.",
            document=Serp,
            index=config.es.index_serps,
        ),
        _get_statistics(
            config=config,
            name="+ URL page",
            description="SERPs for which the page has been parsed from the URL.",
            document=Serp,
            index=config.es.index_serps,
            last_modified_field="url_page_parser.last_parsed",
        ),
        _get_statistics(
            config=config,
            name="+ URL offset",
            description="SERPs for which the offset has been parsed from the URL.",
            document=Serp,
            index=config.es.index_serps,
            last_modified_field="url_offset_parser.last_parsed",
        ),
        _get_statistics(
            config=config,
            name="+ WARC",
            description="SERPs for which the WARC has been downloaded.",
            document=Serp,
            index=config.es.index_serps,
            last_modified_field="warc_downloader.last_downloaded",
        ),
        _get_statistics(
            config=config,
            name="+ WARC query",
            description="SERPs for which the query has been parsed from the WARC.",
            document=Serp,
            index=config.es.index_serps,
            last_modified_field="warc_query_parser.last_parsed",
        ),
        _get_statistics(
            config=config,
            name="+ WARC snippets",
            description="SERPs for which the snippets have been parsed from the WARC.",
            document=Serp,
            index=config.es.index_serps,
            last_modified_field="warc_snippets_parser.last_parsed",
        ),
        _get_warc_cache_statistics(
            config=config,
            name="→ WARC cache (ready)",
            description="Downloaded SERP WARC records, ready to be uploaded to S3.",
            cache_path = config.warc_cache.path_serps,
            temporary=False,
        ),
        _get_warc_cache_statistics(
            config=config,
            name="→ WARC cache (in progress)",
            description="Downloaded SERP WARC records, still locked by a downloader.",
            cache_path = config.warc_cache.path_serps,
            temporary=True,
        ),
        _get_statistics(
            config=config,
            name="Results",
            description="Search result from the SERPs.",
            document=Result,
            index=config.es.index_results,
        ),
        _get_statistics(
            config=config,
            name="URL query parsers",
            description="Parser to get the query from a SERP's URL.",
            document=UrlQueryParser,
            index=config.es.index_url_query_parsers,
        ),
        _get_statistics(
            config=config,
            name="URL page parsers",
            description="Parser to get the page from a SERP's URL.",
            document=UrlPageParser,
            index=config.es.index_url_page_parsers,
        ),
        _get_statistics(
            config=config,
            name="URL offset parsers",
            description="Parser to get the offset from a SERP's URL.",
            document=UrlOffsetParser,
            index=config.es.index_url_offset_parsers,
        ),
        _get_statistics(
            config=config,
            name="WARC query parsers",
            description="Parser to get the query from a SERP's WARC contents.",
            document=WarcQueryParser,
            index=config.es.index_warc_query_parsers,
        ),
        _get_statistics(
            config=config,
            name="WARC snippets parsers",
            description="Parser to get the snippets from a SERP's WARC contents.",
            document=WarcSnippetsParser,
            index=config.es.index_warc_snippets_parsers,
        ),
    ]

    progress_list: list[Progress] = [
        _get_processed_progress(
            config=config,
            input_name="Archives",
            output_name="Sources",
            description="Build sources for all archives.",
            document=Archive,
            index=config.es.index_archives,
            filter_query=~Exists(field="exclusion_reason"),
            status_field="should_build_sources",
        ),
        _get_processed_progress(
            config=config,
            input_name="Providers",
            output_name="Sources",
            description="Build sources for all search providers.",
            document=Provider,
            index=config.es.index_providers,
            filter_query=~Exists(field="exclusion_reason"),
            status_field="should_build_sources",
        ),
        _get_processed_progress(
            config=config,
            input_name="Sources",
            output_name="Captures",
            description="Fetch CDX captures for all domains and "
            "prefixes in the sources.",
            document=Source,
            index=config.es.index_sources,
            status_field="should_fetch_captures",
        ),
        _get_processed_progress(
            config=config,
            input_name="Captures",
            output_name="SERPs",
            description="Parse queries from capture URLs.",
            document=Capture,
            index=config.es.index_captures,
            status_field="url_query_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse page from SERP URLs.",
            document=Serp,
            index=config.es.index_serps,
            status_field="url_page_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse offset from SERP URLs.",
            document=Serp,
            index=config.es.index_serps,
            status_field="url_offset_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Download WARCs.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Term(capture__status_code=200),
            status_field="warc_downloader.should_download",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse query from WARC contents.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="warc_location"),
            status_field="warc_query_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="SERPs",
            output_name="SERPs",
            description="Parse snippets from WARC contents.",
            document=Serp,
            index=config.es.index_serps,
            filter_query=Exists(field="warc_location"),
            status_field="warc_snippets_parser.should_parse",
        ),
        _get_processed_progress(
            config=config,
            input_name="Results",
            output_name="Results",
            description="Download WARCs.",
            document=Result,
            index=config.es.index_results,
            filter_query=Exists(field="snippet.url"),
            status_field="warc_downloader.should_download",
        ),
    ]

    etag = str(
        hash(
            (
                tuple(statistics_list),
                tuple(progress_list),
            )
        )
    )

    response = make_response(
        render_template(
            "home.html",
            statistics_list=statistics_list,
            progress_list=progress_list,
            year=utc_now().year,
        )
    )
    response.headers.add("ETag", etag)
    return response



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_linkedin_vizthink_1229875736():
    verify_serp_parsing(
        "https://web.archive.org/web/20081221160856id_/http://www.linkedin.com:80/search?search=&sortCriteria=3&keywords=VizThink&goback=%2Esrp_1_1211730310191_in",
        "linkedin",
    )


def test_parse_query_linkedin_parent_information_center_of_nj_https_www_linkedin_com_feed_1537815970():
    verify_serp_parsing(
        "https://web.archive.org/web/20180924190610id_/https://www.linkedin.com/search/results/all/?keywords=Parent%20Information%20Center%20of%20NJ/../https://www.linkedin.com/feed/",
        "linkedin",
    )


def test_parse_query_linkedin_g_d_goenka_international_school_surat_wisdom_valley_campus_nr_anuvrat_dwar_new_city_light_road_rcc_canal_road_bar_surat_gujrat_1568400864():
    verify_serp_parsing(
        "https://web.archive.org/web/20190913185424id_/https://www.linkedin.com/search/results/all/?keywords=G.%20D.%20Goenka%20International%20School%20Surat%20Wisdom%20Valley%20campus,%20Nr.%20Anuvrat%20Dwar,%20New%20City%20Light%20Road,%20RCC%20Canal%20Road,%20Bar%20Surat%20GUJRAT",
        "linkedin",
    )


def test_parse_query_linkedin_test_1563279735():
    verify_serp_parsing(
        "https://web.archive.org/web/20190716122215id_/https://www.linkedin.com/search/results/content/?keywords=test&origin=test%20UNION%20ALL%20SELECT%20NULL%2CNULL%2CNULL--%20FPED&",
        "linkedin",
    )


def test_parse_query_linkedin_cryptocurrencies_1556702372():
    verify_serp_parsing(
        "https://web.archive.org/web/20190501091932id_/https://www.linkedin.com/search/results/content/?keywords=%23cryptocurrencies&origin=GLOBAL_SEARCH_HEADER",
        "linkedin",
    )


def test_parse_query_linkedin_james_margolin_fbi_1571382052():
    verify_serp_parsing(
        "https://web.archive.org/web/20191018070052id_/https://www.linkedin.com/search/results/all/?keywords=james%20margolin%20fbi&origin=GLOBAL_SEARCH_HEADER",
        "linkedin",
    )


def test_parse_query_linkedin_parent_information_center_of_nj_https_www_linkedin_com_search_results_all_keywords_parent_information_center_of_nj_1537815969():
    verify_serp_parsing(
        "https://web.archive.org/web/20180924190609id_/https://www.linkedin.com/search/results/all/?keywords=Parent%20Information%20Center%20of%20NJ/../https://www.linkedin.com/search/results/all/?keywords=Parent%20Information%20Center%20of%20NJ",
        "linkedin",
    )



from dataclasses import dataclass
from gzip import GzipFile
from pathlib import Path
from typing import Sized, Iterable, Iterator

from archive_query_log.legacy.model import ArchivedQueryUrl
from archive_query_log.legacy.util.text import count_lines, text_io_wrapper


@dataclass(frozen=True)
class ArchivedQueryUrls(Sized, Iterable[ArchivedQueryUrl]):
    """
    Read archived query URLs from a JSONL file.
    """

    path: Path
    """
    Path where the query URLs are stored in JSONL format.
    """

    def __post_init__(self):
        self._check_urls_path()

    def _check_urls_path(self):
        if not self.path.exists() or not self.path.is_file():
            raise ValueError(
                f"URLs path must be a file: {self.path}"
            )

    def __len__(self) -> int:
        with (self.path.open("rb") as file,
              GzipFile(fileobj=file, mode="rb") as gzip_file):
            return count_lines(gzip_file)

    def __iter__(self) -> Iterator[ArchivedQueryUrl]:
        schema = ArchivedQueryUrl.schema()
        with (self.path.open("rb") as file,
              GzipFile(fileobj=file, mode="rb") as gzip_file,
              text_io_wrapper(gzip_file) as text_file):
            for line in text_file:
                query_url = schema.loads(line, many=True)
                if isinstance(query_url, list):
                    raise ValueError(
                        f"Expected one query URL per line: {line}")
                yield query_url



from urllib.parse import urljoin

from click import echo
from tqdm.auto import tqdm

from archive_query_log.archives import add_archive
from archive_query_log.config import Config

_ARCHIVE_IT_METADATA_FIELDS = [
    "Title",
    "Description",
    "Subject",
    "Coverage",
    "Language",
    "Collector",
    "Creator",
    "Publisher",
    "Date",
    "Identifier",
    "Rights",
]

DEFAULT_ARCHIVE_IT_API_URL: str = "https://partner.archive-it.org"
DEFAULT_ARCHIVE_IT_WAYBACK_URL: str = "https://wayback.archive-it.org/"
DEFAULT_ARCHIVE_IT_PAGE_SIZE: int = 100


def import_archives(
        config: Config,
        api_url: str = DEFAULT_ARCHIVE_IT_API_URL,
        wayback_url: str = DEFAULT_ARCHIVE_IT_WAYBACK_URL,
        page_size: int = DEFAULT_ARCHIVE_IT_PAGE_SIZE,
        priority: float | None = None,
        no_merge: bool = False,
        auto_merge: bool = False,
) -> None:
    echo("Load Archive-It collections.")
    collections_api_url = urljoin(api_url, "/api/collection")
    response = config.http.session.get(
        collections_api_url,
        params=[
            ("limit", 0),
            ("format", "json"),
        ],
    )
    num_collections = int(response.headers["Total-Row-Count"])
    echo(f"Found {num_collections} collections on Archive-It.")

    # noinspection PyTypeChecker
    progress = tqdm(total=num_collections, desc="Import archives",
                    unit="archives", disable=not auto_merge and not no_merge)
    offset_range = range(0, num_collections, page_size)
    for offset in offset_range:
        response = config.http.session.get(
            collections_api_url,
            params=[
                ("limit", page_size),
                ("offset", offset),
                ("format", "json"),
            ],
        )
        response_list = response.json()
        for item in response_list:
            name = f"Archive-It {item['name']}"
            archive_it_id = int(item["id"])

            description_parts = []
            metadata = item["metadata"]
            for metadata_field in _ARCHIVE_IT_METADATA_FIELDS:
                if metadata_field in metadata:
                    for title in metadata[metadata_field]:
                        description_parts.append(
                            f"{metadata_field}: {title['value']}")
            description_parts.append(f"Archive-It ID: {archive_it_id}")
            description = "\n".join(description_parts)
            cdx_api_url = urljoin(
                wayback_url, f"{archive_it_id}/timemap/cdx")
            memento_api_url = urljoin(wayback_url, f"{archive_it_id}")
            add_archive(
                config=config,
                name=name,
                description=description,
                cdx_api_url=cdx_api_url,
                memento_api_url=memento_api_url,
                priority=priority,
                no_merge=no_merge,
                auto_merge=auto_merge,
            )
            progress.update(1)



from dataclasses import dataclass
from typing import Pattern, Iterator
from urllib.parse import urljoin

from bs4 import Tag

from archive_query_log.legacy.model import ArchivedSearchResultSnippet, \
    HighlightedText
from archive_query_log.legacy.results.parse import HtmlResultsParser
from archive_query_log.legacy.util.html import clean_html


@dataclass(frozen=True)
class ChatNoirResultsParser(HtmlResultsParser):
    url_pattern: Pattern[str]

    def parse_html(
            self,
            html: Tag,
            timestamp: int,
            serp_url: str,
    ) -> Iterator[ArchivedSearchResultSnippet]:
        results = html.find("section", id="SearchResults")
        if results is None or not isinstance(results, Tag):
            return
        results_iter = results.find_all("article", class_="search-result")
        for index, result in enumerate(results_iter):
            header = result.find("header")
            url: str = header.find("a", class_="link")["href"]
            url = urljoin(serp_url, url)
            title = HighlightedText(clean_html(header.find("h2")))
            # Remove header. Only the snippet will be left.
            header.decompose()
            snippet = HighlightedText(clean_html(result))
            yield ArchivedSearchResultSnippet(
                rank=index + 1,
                url=url,
                timestamp=timestamp,
                title=title,
                snippet=snippet if len(snippet) > 0 else None,
            )



# flake8: noqa
# This file is auto-generated by generate_tests.py.
from archive_query_log.legacy.results.test.test_utils import verify_serp_parsing


def test_parse_query_qwant_administrateur_general_du_cnam_1619206522():
    verify_serp_parsing(
        "https://web.archive.org/web/20210423193522id_/https://www.qwant.com/?q=%22Administrateur+g%C3%A9n%C3%A9ral+du+Cnam%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_great_falls_gun_club_1637348184():
    verify_serp_parsing(
        "https://web.archive.org/web/20211119185624id_/https://www.qwant.com/?q=great+falls+gun+club&client=ext-firefox-sb&t=web",
        "qwant",
    )


def test_parse_query_qwant_rosa_linux_1664389965():
    verify_serp_parsing(
        "https://web.archive.org/web/20220928183245id_/https://www.qwant.com/?q=%22ROSA+Linux%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_the_owl_club_1666374443():
    verify_serp_parsing(
        "https://web.archive.org/web/20221021174723id_/https://www.qwant.com/?q=the+owl+club&client=ext-firefox-sb&t=web",
        "qwant",
    )


def test_parse_query_qwant_journee_nationale_du_drapeau_gabonais_1628405454():
    verify_serp_parsing(
        "https://web.archive.org/web/20210808065054id_/https://www.qwant.com/?q=%22Journ%C3%A9e+nationale+du+drapeau+gabonais%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_drapeau_canadien_1615473541():
    verify_serp_parsing(
        "https://web.archive.org/web/20210311143901id_/https://www.qwant.com/?q=drapeau+canadien&client=brz-brave",
        "qwant",
    )


def test_parse_query_qwant_hovelhof_1461470750():
    verify_serp_parsing(
        "https://web.archive.org/web/20160424040550id_/https://www.qwant.com/?q=h%C3%B6velhof",
        "qwant",
    )


def test_parse_query_qwant_eric_singer_drums_1621083357():
    verify_serp_parsing(
        "https://web.archive.org/web/20210515125557id_/https://www.qwant.com/?q=eric+singer+drums&t=all",
        "qwant",
    )


def test_parse_query_qwant_obs_keyframe_interval_1657322209():
    verify_serp_parsing(
        "https://web.archive.org/web/20220708231649id_/https://www.qwant.com/?q=obs+keyframe+interval&client=ext-firefox-sb&t=web",
        "qwant",
    )


def test_parse_query_qwant_celine_de_roany_1640557696():
    verify_serp_parsing(
        "https://web.archive.org/web/20211226222816id_/https://www.qwant.com/?q=%22C%C3%A9line+de+Roany%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_festival_corazon_latino_1644332358():
    verify_serp_parsing(
        "https://web.archive.org/web/20220208145918id_/https://www.qwant.com/?q=%22Festival+Coraz%C3%B3n+latino%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_trouve_arc_bf4_1625314405():
    verify_serp_parsing(
        "https://web.archive.org/web/20210703121325id_/https://www.qwant.com/?q=trouv%C3%A9+arc+bf4&client=brz-brave",
        "qwant",
    )


def test_parse_query_qwant_dr_heim_tampa_1631748962():
    verify_serp_parsing(
        "https://web.archive.org/web/20210915233602id_/https://www.qwant.com/?q=dr+heim+tampa&t=all",
        "qwant",
    )


def test_parse_query_qwant_sacred_age_odysseria_1652523638():
    verify_serp_parsing(
        "https://web.archive.org/web/20220514102038id_/https://www.qwant.com/?q=%2BSacred+Age+Odysseria&client=ext-firefox-sb&t=web",
        "qwant",
    )


def test_parse_query_qwant_musique_folklorique_bretonne_1631820741():
    verify_serp_parsing(
        "https://web.archive.org/web/20210916193221id_/https://www.qwant.com/?q=musique+folklorique+bretonne&t=all",
        "qwant",
    )


def test_parse_query_qwant_initialiser_debat_dadmissibilite_1645273237():
    verify_serp_parsing(
        "https://web.archive.org/web/20220219122037id_/https://www.qwant.com/?q=%22Initialiser+D%C3%A9bat+d%26%2339%3Badmissibilit%C3%A9%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_geographie_de_la_france_1620931197():
    verify_serp_parsing(
        "https://web.archive.org/web/20210513183957id_/https://www.qwant.com/?q=geographie+de+la+france&t=all",
        "qwant",
    )


def test_parse_query_qwant_anna_sottilotta_1630417705():
    verify_serp_parsing(
        "https://web.archive.org/web/20210831134825id_/https://www.qwant.com/?q=anna+sottilotta&t=all",
        "qwant",
    )


def test_parse_query_qwant_metal_slug_foo_1624572900():
    verify_serp_parsing(
        "https://web.archive.org/web/20210624221500id_/https://www.qwant.com/?q=metal+slug+foo&client=brz-brave",
        "qwant",
    )


def test_parse_query_qwant_another_word_for_less_likely_1641721403():
    verify_serp_parsing(
        "https://web.archive.org/web/20220109094323id_/https://www.qwant.com/?q=another+word+for+less+likely&t=all",
        "qwant",
    )


def test_parse_query_qwant_rmvmn_1619293679():
    verify_serp_parsing(
        "https://web.archive.org/web/20210424194759id_/https://www.qwant.com/?q=%22RMVMN%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_unification_album_1626264528():
    verify_serp_parsing(
        "https://web.archive.org/web/20210714120848id_/https://www.qwant.com/?q=%22Unification+(album)%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_rob_diamond_movies_1626693763():
    verify_serp_parsing(
        "https://web.archive.org/web/20210719112243id_/https://www.qwant.com/?q=rob+diamond+movies&t=all",
        "qwant",
    )


def test_parse_query_qwant_code_9_1661507082():
    verify_serp_parsing(
        "https://web.archive.org/web/20220826094442id_/https://www.qwant.com/?q=%22Code+9%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_bluestacks_keyboard_not_working_1637463746():
    verify_serp_parsing(
        "https://web.archive.org/web/20211121030226id_/https://www.qwant.com/?q=bluestacks+keyboard+not+working&client=brz-brave&t=web",
        "qwant",
    )


def test_parse_query_qwant_jesus_mendez_facebook_1642962807():
    verify_serp_parsing(
        "https://web.archive.org/web/20220123183327id_/https://www.qwant.com/?q=jesus+mendez+facebook&t=all",
        "qwant",
    )


def test_parse_query_qwant_marcel_caya_1628794563():
    verify_serp_parsing(
        "https://web.archive.org/web/20210812185603id_/https://www.qwant.com/?q=%22Marcel+Caya%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_david_shepard_restaurateur_de_film_1595246740():
    verify_serp_parsing(
        "https://web.archive.org/web/20200720120540id_/https://www.qwant.com/?q=%22David+Shepard+(restaurateur+de+film)%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_phantasialand_1461344566():
    verify_serp_parsing(
        "https://web.archive.org/web/20160422170246id_/https://www.qwant.com/?q=phantasialand",
        "qwant",
    )


def test_parse_query_qwant_attaque_a_tunis_1433356632():
    verify_serp_parsing(
        "https://web.archive.org/web/20150603183712id_/https://www.qwant.com/?q=Attaque+%C3%A0+Tunis&source=news",
        "qwant",
    )


def test_parse_query_qwant_neonazis_en_france_1663085614():
    verify_serp_parsing(
        "https://web.archive.org/web/20220913161334id_/https://www.qwant.com/?client=brz-moz&t=videos&q=n%C3%A9onazis+en+France&o=0%3AMJrfVZ7PGdc",
        "qwant",
    )


def test_parse_query_qwant_victoria_zdrok_1632510384():
    verify_serp_parsing(
        "https://web.archive.org/web/20210924190624id_/https://www.qwant.com/?q=%22Victoria+Zdrok%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_ouat_once_upon_a_train_1643380520():
    verify_serp_parsing(
        "https://web.archive.org/web/20220128143520id_/https://www.qwant.com/?q=%22OUAT+-+Once+Upon+A+Train%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_skerdi_dervishaj_1635917093():
    verify_serp_parsing(
        "https://web.archive.org/web/20211103052453id_/https://www.qwant.com/?q=%22Skerdi+Dervishaj%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_allades_1631959376():
    verify_serp_parsing(
        "https://web.archive.org/web/20210918100256id_/https://www.qwant.com/?q=%22Allad%C3%A8s%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_world_deutsch_regional_europa_deutschland_baden_wurttemberg_landkreise_heilbronn_stadte_und_gemeinde_1626329085():
    verify_serp_parsing(
        "https://web.archive.org/web/20210715060445id_/https://www.qwant.com/?q=World+Deutsch+Regional+Europa+Deutschland+Baden-W%C3%BCrttemberg+Landkreise+Heilbronn+St%C3%A4dte+und+Gemeinden+W%C3%BCstenrot",
        "qwant",
    )


def test_parse_query_qwant_toddlercon_1579361594():
    verify_serp_parsing(
        "https://web.archive.org/web/20200118153314id_/https://www.qwant.com/?client=brz-vivaldi&q=Toddlercon",
        "qwant",
    )


def test_parse_query_qwant_guerre_froide_1654846351():
    verify_serp_parsing(
        "https://web.archive.org/web/20220610073231id_/https://www.qwant.com/?l=fr&t=videos&q=guerre+froide",
        "qwant",
    )


def test_parse_query_qwant_lombocruralgie_francais_1656441384():
    verify_serp_parsing(
        "https://web.archive.org/web/20220628183624id_/https://www.qwant.com/?q=lombocruralgie+francais&t=all",
        "qwant",
    )


def test_parse_query_qwant_volt_bulgarie_1628499772():
    verify_serp_parsing(
        "https://web.archive.org/web/20210809090252id_/https://www.qwant.com/?q=%22Volt+Bulgarie%22&t=all",
        "qwant",
    )


def test_parse_query_qwant_what_is_beowulf_about_summary_1659304409():
    verify_serp_parsing(
        "https://web.archive.org/web/20220731215329id_/https://www.qwant.com/?q=what+is+beowulf+about+summary&client=ext-chrome-sb&t=web",
        "qwant",
    )


def test_parse_query_qwant_tv_horror_movies_chart_1667140806():
    verify_serp_parsing(
        "https://web.archive.org/web/20221030144006id_/https://www.qwant.com/?q=%2Ftv%2F+horror+movies+chart&client=opensearch&t=web",
        "qwant",
    )


def test_parse_query_qwant_steam_ubuntu_wiki_1592725715():
    verify_serp_parsing(
        "https://web.archive.org/web/20200621074835id_/https://www.qwant.com/?client=brz-moz&q=Steam+ubuntu+wiki",
        "qwant",
    )


def test_parse_query_qwant_que_tal_magazine_1643474954():
    verify_serp_parsing(
        "https://web.archive.org/web/20220129164914id_/https://www.qwant.com/?q=que+tal+magazine&t=all",
        "qwant",
    )


def test_parse_query_qwant_sporting_stadium_1640362241():
    verify_serp_parsing(
        "https://web.archive.org/web/20211224161041id_/https://www.qwant.com/?q=sporting+stadium&t=all",
        "qwant",
    )


def test_parse_query_qwant_danse_1460250714():
    verify_serp_parsing(
        "https://web.archive.org/web/20160410011154id_/https://www.qwant.com/?q=danse",
        "qwant",
    )


def test_parse_query_qwant_lemke_software_gmbh_1625227146():
    verify_serp_parsing(
        "https://web.archive.org/web/20210702115906id_/https://www.qwant.com/?q=lemke+software+gmbh&t=all",
        "qwant",
    )


def test_parse_query_qwant_dr_michael_valliere_1653406352():
    verify_serp_parsing(
        "https://web.archive.org/web/20220524153232id_/https://www.qwant.com/?q=dr+michael+valliere&t=all",
        "qwant",
    )


def test_parse_query_qwant_fredrik_lundberg_1665847606():
    verify_serp_parsing(
        "https://web.archive.org/web/20221015152646id_/https://www.qwant.com/?q=fredrik+lundberg&client=opensearch&t=web",
        "qwant",
    )


def test_parse_query_qwant_i_squared_capital_linkedin_1653657436():
    verify_serp_parsing(
        "https://web.archive.org/web/20220527131716id_/https://www.qwant.com/?q=i+squared+capital+linkedin&t=all",
        "qwant",
    )



from pathlib import Path

from flask import Flask
from werkzeug import run_simple

from archive_query_log import __name__ as app_name
from archive_query_log.config import Config
from archive_query_log.monitoring.home import home


def monitoring_app(config: Config) -> Flask:
    app = Flask(app_name)
    app.add_url_rule("/", "home", lambda: home(config))
    return app


def run_monitoring(config: Config, host: str, port: int) -> None:
    app = monitoring_app(config)
    if app.template_folder is None:
        template_file_names = []
    else:
        template_dir_path: Path = Path(app.root_path) / app.template_folder
        template_file_paths = [
            template_dir_path / template
            for template in app.jinja_env.list_templates()
        ]
        template_file_names = [
            str(template) for template in template_file_paths
        ]
    run_simple(
        hostname=host,
        port=port,
        application=app,
        use_reloader=True,
        use_debugger=True,
        extra_files=template_file_names,
    )



from click import group, option, IntRange, FloatRange

from archive_query_log.cli.util import pass_config
from archive_query_log.config import Config
from archive_query_log.imports.archive_it import \
    DEFAULT_ARCHIVE_IT_PAGE_SIZE, DEFAULT_ARCHIVE_IT_WAYBACK_URL, \
    DEFAULT_ARCHIVE_IT_API_URL
from archive_query_log.orm import Archive


@group()
def archives() -> None:
    pass


@archives.command()
@option("-n", "--name", type=str, required=True,
        prompt="Name")
@option("-d", "--description", type=str)
@option("-c", "--cdx-api-url", type=str, required=True,
        prompt="CDX API URL", metavar="URL")
@option("-m", "--memento-api-url", type=str, required=True,
        prompt="Memento API URL", metavar="URL")
@option("--priority", type=FloatRange(min=0, min_open=False))
@pass_config
def add(
        config: Config,
        name: str,
        description: str | None,
        cdx_api_url: str,
        memento_api_url: str,
        priority: float | None,
) -> None:
    from archive_query_log.archives import add_archive
    Archive.init(using=config.es.client, index=config.es.index_archives)
    add_archive(
        config=config,
        name=name,
        description=description,
        cdx_api_url=cdx_api_url,
        memento_api_url=memento_api_url,
        priority=priority,
    )


@archives.group("import")
def import_() -> None:
    pass


@import_.command()
@option("--api-url", type=str, required=True,
        default=DEFAULT_ARCHIVE_IT_API_URL, metavar="URL")
@option("--wayback-url", type=str, required=True,
        default=DEFAULT_ARCHIVE_IT_WAYBACK_URL, metavar="URL")
@option("--page-size", type=IntRange(min=1), required=True,
        default=DEFAULT_ARCHIVE_IT_PAGE_SIZE)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--no-merge", is_flag=True, default=False, type=bool)
@option("--auto-merge", is_flag=True, default=False, type=bool)
@pass_config
def archive_it(
        config: Config,
        api_url: str,
        wayback_url: str,
        page_size: int,
        priority: float | None,
        no_merge: bool,
        auto_merge: bool,
) -> None:
    from archive_query_log.imports.archive_it import import_archives
    Archive.init(using=config.es.client, index=config.es.index_archives)
    import_archives(
        config=config,
        api_url=api_url,
        wayback_url=wayback_url,
        page_size=page_size,
        no_merge=no_merge,
        auto_merge=auto_merge,
        priority=priority,
    )



from pathlib import Path

from click import group, option, Choice, Path as PathType, UsageError, \
    FloatRange

from archive_query_log.cli.util import pass_config
from archive_query_log.config import Config
from archive_query_log.orm import UrlQueryParserType, \
    UrlQueryParser, UrlPageParserType, UrlPageParser, \
    UrlOffsetParser, UrlOffsetParserType, WarcQueryParserType, \
    WarcQueryParser, WarcSnippetsParserType, WarcSnippetsParser, \
    WarcDirectAnswersParserType, WarcDirectAnswersParser, \
    WarcMainContentParserType, WarcMainContentParser


@group()
def parsers() -> None:
    pass


@parsers.group()
def url_query() -> None:
    pass


CHOICES_URL_QUERY_PARSER_TYPE = [
    "query-parameter",
    "fragment-parameter",
    "path-segment",
]


@url_query.command("add")
@option("--provider-id", type=str)
@option("--url-pattern-regex", type=str)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--parser-type",
        type=Choice(CHOICES_URL_QUERY_PARSER_TYPE), required=True)
@option("--parameter", type=str)
@option("--segment", type=int)
@option("--remove-pattern-regex", type=str)
@option("--space-pattern-regex", type=str)
@pass_config
def url_query_add(
        config: Config,
        provider_id: str | None,
        url_pattern_regex: str | None,
        priority: float | None,
        parser_type: str,
        parameter: str | None,
        segment: int | None,
        remove_pattern_regex: str | None,
        space_pattern_regex: str | None,
) -> None:
    from archive_query_log.parsers.url_query import add_url_query_parser
    parser_type_strict: UrlQueryParserType
    if parser_type == "query-parameter":
        parser_type_strict = "query_parameter"
        if parameter is None:
            raise UsageError("No query parameter given.")
    elif parser_type == "fragment-parameter":
        parser_type_strict = "fragment_parameter"
        if parameter is not None:
            raise UsageError("No fragment parameter given.")
    elif parser_type == "path-segment":
        parser_type_strict = "path_segment"
        if segment is None:
            raise UsageError("No path segment given.")
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    UrlQueryParser.init(using=config.es.client, index=config.es.index_url_query_parsers)
    add_url_query_parser(
        config=config,
        provider_id=provider_id,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type_strict,
        parameter=parameter,
        segment=segment,
        remove_pattern_regex=remove_pattern_regex,
        space_pattern_regex=space_pattern_regex,
    )


@url_query.command("import")
@option("-s", "--services-file", "services_path",
        type=PathType(path_type=Path, exists=True, file_okay=True,
                      dir_okay=False, readable=True, resolve_path=True,
                      allow_dash=False),
        default=Path("data") / "selected-services.yaml")
@pass_config
def url_query_import(config: Config, services_path: Path) -> None:
    from archive_query_log.imports.yaml import import_url_query_parsers
    UrlQueryParser.init(using=config.es.client, index=config.es.index_url_query_parsers)
    import_url_query_parsers(config, services_path)


@parsers.group()
def url_page() -> None:
    pass


CHOICES_URL_PAGE_PARSER_TYPE = [
    "query-parameter",
    "fragment-parameter",
    "path-segment",
]


@url_page.command("add")
@option("--provider-id", type=str)
@option("--url-pattern-regex", type=str)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--parser-type",
        type=Choice(CHOICES_URL_PAGE_PARSER_TYPE), required=True)
@option("--parameter", type=str)
@option("--segment", type=int)
@option("--remove-pattern-regex", type=str)
@option("--space-pattern-regex", type=str)
@pass_config
def url_page_add(
        config: Config,
        provider_id: str | None,
        url_pattern_regex: str | None,
        priority: float | None,
        parser_type: str,
        parameter: str | None,
        segment: int | None,
        remove_pattern_regex: str | None,
        space_pattern_regex: str | None,
) -> None:
    from archive_query_log.parsers.url_page import add_url_page_parser
    parser_type_strict: UrlPageParserType
    if parser_type == "query-parameter":
        parser_type_strict = "query_parameter"
        if parameter is None:
            raise UsageError("No query parameter given.")
    elif parser_type == "fragment-parameter":
        parser_type_strict = "fragment_parameter"
        if parameter is not None:
            raise UsageError("No fragment parameter given.")
    elif parser_type == "path-segment":
        parser_type_strict = "path_segment"
        if segment is None:
            raise UsageError("No path segment given.")
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    UrlPageParser.init(using=config.es.client, index=config.es.index_url_page_parsers)
    add_url_page_parser(
        config=config,
        provider_id=provider_id,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type_strict,
        parameter=parameter,
        segment=segment,
        remove_pattern_regex=remove_pattern_regex,
        space_pattern_regex=space_pattern_regex,
    )


@url_page.command("import")
@option("-s", "--services-file", "services_path",
        type=PathType(path_type=Path, exists=True, file_okay=True,
                      dir_okay=False, readable=True, resolve_path=True,
                      allow_dash=False),
        default=Path("data") / "selected-services.yaml")
@pass_config
def url_page_import(config: Config, services_path: Path) -> None:
    from archive_query_log.imports.yaml import import_url_page_parsers
    UrlPageParser.init(using=config.es.client, index=config.es.index_url_page_parsers)
    import_url_page_parsers(config, services_path)


@parsers.group()
def url_offset() -> None:
    pass


CHOICES_URL_OFFSET_PARSER_TYPE = [
    "query-parameter",
    "fragment-parameter",
    "path-segment",
]


@url_offset.command("add")
@option("--provider-id", type=str)
@option("--url-pattern-regex", type=str)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--parser-type",
        type=Choice(CHOICES_URL_OFFSET_PARSER_TYPE), required=True)
@option("--parameter", type=str)
@option("--segment", type=int)
@option("--remove-pattern-regex", type=str)
@option("--space-pattern-regex", type=str)
@pass_config
def url_offset_add(
        config: Config,
        provider_id: str | None,
        url_pattern_regex: str | None,
        priority: float | None,
        parser_type: str,
        parameter: str | None,
        segment: int | None,
        remove_pattern_regex: str | None,
        space_pattern_regex: str | None,
) -> None:
    from archive_query_log.parsers.url_offset import add_url_offset_parser
    parser_type_strict: UrlOffsetParserType
    if parser_type == "query-parameter":
        parser_type_strict = "query_parameter"
        if parameter is None:
            raise UsageError("No query parameter given.")
    elif parser_type == "fragment-parameter":
        parser_type_strict = "fragment_parameter"
        if parameter is not None:
            raise UsageError("No fragment parameter given.")
    elif parser_type == "path-segment":
        parser_type_strict = "path_segment"
        if segment is None:
            raise UsageError("No path segment given.")
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    UrlOffsetParser.init(using=config.es.client, index=config.es.index_url_offset_parsers)
    add_url_offset_parser(
        config=config,
        provider_id=provider_id,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type_strict,
        parameter=parameter,
        segment=segment,
        remove_pattern_regex=remove_pattern_regex,
        space_pattern_regex=space_pattern_regex,
    )


@url_offset.command("import")
@option("-s", "--services-file", "services_path",
        type=PathType(path_type=Path, exists=True, file_okay=True,
                      dir_okay=False, readable=True, resolve_path=True,
                      allow_dash=False),
        default=Path("data") / "selected-services.yaml")
@pass_config
def url_offset_import(config: Config, services_path: Path) -> None:
    from archive_query_log.imports.yaml import import_url_offset_parsers
    UrlOffsetParser.init(using=config.es.client, index=config.es.index_url_offset_parsers)
    import_url_offset_parsers(config, services_path)


@parsers.group()
def warc_query() -> None:
    pass


CHOICES_WARC_QUERY_PARSER_TYPE = [
    "xpath",
]


@warc_query.command("add")
@option("--provider-id", type=str)
@option("--url-pattern-regex", type=str)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--parser-type",
        type=Choice(CHOICES_WARC_QUERY_PARSER_TYPE), required=True)
@option("--xpath", type=str)
@option("--remove-pattern-regex", type=str)
@option("--space-pattern-regex", type=str)
@pass_config
def warc_query_add(
        config: Config,
        provider_id: str | None,
        url_pattern_regex: str | None,
        priority: float | None,
        parser_type: str,
        xpath: str | None,
        remove_pattern_regex: str | None,
        space_pattern_regex: str | None,
) -> None:
    from archive_query_log.parsers.warc_query import add_warc_query_parser
    parser_type_strict: WarcQueryParserType
    if parser_type == "xpath":
        parser_type_strict = "xpath"
        if xpath is None:
            raise UsageError("No XPath given.")
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    WarcQueryParser.init(using=config.es.client, index=config.es.index_warc_query_parsers)
    add_warc_query_parser(
        config=config,
        provider_id=provider_id,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type_strict,
        xpath=xpath,
        remove_pattern_regex=remove_pattern_regex,
        space_pattern_regex=space_pattern_regex,
    )


@warc_query.command("import")
@option("-s", "--services-file", "services_path",
        type=PathType(path_type=Path, exists=True, file_okay=True,
                      dir_okay=False, readable=True, resolve_path=True,
                      allow_dash=False),
        default=Path("data") / "selected-services.yaml")
@pass_config
def warc_query_import(config: Config, services_path: Path) -> None:
    from archive_query_log.imports.yaml import import_warc_query_parsers
    WarcQueryParser.init(using=config.es.client, index=config.es.index_warc_query_parsers)
    import_warc_query_parsers(config, services_path)


@parsers.group()
def warc_snippets() -> None:
    pass


CHOICES_WARC_SNIPPETS_PARSER_TYPE = [
    "xpath",
]


@warc_snippets.command("add")
@option("--provider-id", type=str)
@option("--url-pattern-regex", type=str)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--parser-type",
        type=Choice(CHOICES_WARC_SNIPPETS_PARSER_TYPE), required=True)
@option("--xpath", type=str)
@option("--url-xpath", type=str)
@option("--title-xpath", type=str)
@option("--text-xpath", type=str)
@pass_config
def warc_snippets_add(
        config: Config,
        provider_id: str | None,
        url_pattern_regex: str | None,
        priority: float | None,
        parser_type: str,
        xpath: str | None,
        url_xpath: str | None,
        title_xpath: str | None,
        text_xpath: str | None,
) -> None:
    from archive_query_log.parsers.warc_snippets import \
        add_warc_snippets_parser
    parser_type_strict: WarcSnippetsParserType
    if parser_type == "xpath":
        parser_type_strict = "xpath"
        if xpath is None:
            raise UsageError("No XPath given.")
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    WarcSnippetsParser.init(using=config.es.client, index=config.es.index_warc_snippets_parsers)
    add_warc_snippets_parser(
        config=config,
        provider_id=provider_id,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type_strict,
        xpath=xpath,
        url_xpath=url_xpath,
        title_xpath=title_xpath,
        text_xpath=text_xpath,
    )


@warc_snippets.command("import")
@option("-s", "--services-file", "services_path",
        type=PathType(path_type=Path, exists=True, file_okay=True,
                      dir_okay=False, readable=True, resolve_path=True,
                      allow_dash=False),
        default=Path("data") / "selected-services.yaml")
@pass_config
def warc_snippets_import(config: Config, services_path: Path) -> None:
    from archive_query_log.imports.yaml import import_warc_snippets_parsers
    WarcSnippetsParser.init(using=config.es.client, index=config.es.index_warc_snippets_parsers)
    import_warc_snippets_parsers(config, services_path)


@parsers.group()
def warc_direct_answers() -> None:
    pass


CHOICES_WARC_DIRECT_ANSWERS_PARSER_TYPE = [
    "xpath",
]


@warc_direct_answers.command("add")
@option("--provider-id", type=str)
@option("--url-pattern-regex", type=str)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--parser-type",
        type=Choice(CHOICES_WARC_DIRECT_ANSWERS_PARSER_TYPE), required=True)
@option("--xpath", type=str)
@option("--url-xpath", type=str)
@option("--text-xpath", type=str)
@pass_config
def warc_direct_answers_add(
        config: Config,
        provider_id: str | None,
        url_pattern_regex: str | None,
        priority: float | None,
        parser_type: str,
        xpath: str | None,
        url_xpath: str | None,
        text_xpath: str | None,
) -> None:
    from archive_query_log.parsers.warc_direct_answers import \
        add_warc_direct_answers_parser
    parser_type_strict: WarcDirectAnswersParserType
    if parser_type == "xpath":
        parser_type_strict = "xpath"
        if xpath is None:
            raise UsageError("No XPath given.")
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    WarcDirectAnswersParser.init(using=config.es.client, index=config.es.index_warc_direct_answers_parsers)
    add_warc_direct_answers_parser(
        config=config,
        provider_id=provider_id,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type_strict,
        xpath=xpath,
        url_xpath=url_xpath,
        text_xpath=text_xpath,
    )


@parsers.group()
def warc_main_content() -> None:
    pass


CHOICES_WARC_MAIN_CONTENT_PARSER_TYPE = [
    "resiliparse",
]


@warc_main_content.command("add")
@option("--provider-id", type=str)
@option("--url-pattern-regex", type=str)
@option("--priority", type=FloatRange(min=0, min_open=False))
@option("--parser-type",
        type=Choice(CHOICES_WARC_MAIN_CONTENT_PARSER_TYPE), required=True)
@pass_config
def warc_main_content_add(
        config: Config,
        provider_id: str | None,
        url_pattern_regex: str | None,
        priority: float | None,
        parser_type: str,
) -> None:
    from archive_query_log.parsers.warc_main_content import \
        add_warc_main_content_parser
    parser_type_strict: WarcMainContentParserType
    if parser_type == "resiliparse":
        parser_type_strict = "resiliparse"
    else:
        raise ValueError(f"Invalid parser type: {parser_type}")
    WarcMainContentParser.init(using=config.es.client, index=config.es.index_warc_direct_answers_parsers)
    add_warc_main_content_parser(
        config=config,
        provider_id=provider_id,
        url_pattern_regex=url_pattern_regex,
        priority=priority,
        parser_type=parser_type_strict,
    )
